; Versuch eines Tutorials: Benutzung von Assembler-Befehlen in PureBasic-Programmen, 32-Bit-Windows
; Bezieht sich auf FAsm, da dieser auch in PB verwendet wird 
; "Helle" Klaus Helbing, 19.07.2011, PB v4.51 (x86) und PB 4.60 Beta 3 (x86)

;=========================================================================================

; VORBEMERKUNGEN
; Wozu Assembler in PB? Weil damit der volle Leistungsumfang des Prozessors 
; ausgenutzt werden kann! CPU-Befehle, die PB noch nicht untersttzt, knnen so
; z.B. fr auf Geschwindigkeit optimierte Programme genutzt werden. Es lassen
; sich auch spezielle Optimierungen durchfhren: Ein (allgemeiner) PB-Befehl 
; muss universell sein und alle mglichen Aufruf-Varianten fehlerfrei abdecken.
; Dabei checkt er z.B. auch zeitraubend Sachen, die fr das gerade zu schreibende
; Programm nicht ntig wren. Mit Assembler ist es mglich, speziell angepassten
; Code zu erstellen.       
; Von PB angegebene Einschrnkungen zum Gebrauch der Prozessor-Register brauchen
; uns nicht abzuschrecken; mit der ntigen Sorgfalt sind ALLE uns interessierenden
; Register ohne Wenn und Aber nutzbar.

; FPU- und SIMD-Befehle sind nicht Inhalt dieses Tutorials. 

;=========================================================================================

; HARDWARE-VORAUSSETZUNGEN
; Wie fr PB (Windows) auch; also mindestens 32-Bit-CPU und 32-Bit-Windows-Betriebssystem.

;=========================================================================================

; TOOLS
; Nachfolgendes Macro kann benutzt werden um sich die wichtigsten Register-Werte und Flag-
; Zustnde in der Window-Titel-Leiste anzeigen zu lassen: 
;-------- 
Macro asminfo
Global VarAX.l          ;der Ordnung halber alle Variablen deklarieren
Global VarBX.l
Global VarCX.l
Global VarDX.l
Global VarDI.l
Global VarSI.l
Global VarBP.l
Global VarSP.l
Global FA.b             ;C, P, A, Z, S, D, O - Flags
Global FC.b
Global FD.b
Global FO.b
Global FP.b
Global FS.b
Global FZ.b
!PUSHFD                 ;Flags sichern
!PUSHAD                 ;alle Register sichern
!MOV [v_VarAX],eax      ;alle Register in Variablen schreiben
!MOV [v_VarBX],ebx
!MOV [v_VarCX],ecx
!MOV [v_VarDX],edx
!MOV [v_VarDI],edi
!MOV [v_VarSI],esi
!MOV [v_VarBP],ebp
!MOV [v_VarSP],esp

!PUSHFD                 ;das 32-bittige E-FLAG-Register auf den Stack schieben
!POP eax                ;und in das EAX-Register laden

!MOV [v_FC],0           ;fr Carry-Flag; Variable erstmal auf Null
!SHR eax,1              ;nach Rechts shiften, das letzte geshiftete Bit steht im Carry-Flag
!ADC [v_FC],0           ;Wert des Carry-Flags zum Variablen-Wert addieren

!MOV [v_FP],0           ;fr Parity-Flag
!SHR eax,2
!ADC [v_FP],0

!MOV [v_FA],0           ;fr Auxiliary-Flag
!SHR eax,2
!ADC [v_FA],0

!MOV [v_FZ],0           ;fr Zero-Flag
!SHR eax,2
!ADC [v_FZ],0

!MOV [v_FS],0           ;fr Signum-Flag
!SHR eax,1
!ADC [v_FS],0

!MOV [v_FD],0           ;fr Direction-Flag
!SHR eax,3
!ADC [v_FD],0

!MOV [v_FO],0           ;fr Overflow-Flag
!SHR eax,1
!ADC [v_FO],0

Delay(1)                ;sonst funktionierts direkt nach einem MessageRequester nicht
hWnd=GetForegroundWindow_()
String$="EAX="+Hex(VarAX)+"  EBX="+Hex(VarBX)+"  ECX="+Hex(VarCX)+"  EDX="+Hex(VarDX)+"  EDI="+Hex(VarDI)+"  ESI="+Hex(VarSI)+"  EBP="+Hex(VarBP)+"  ESP="+Hex(VarSP)+"  C="+Str(FC)+"  P="+Str(FP)+"  A="+Str(FA)+"  Z="+Str(FZ)+"  S="+Str(FS)+"  D="+Str(FD)+"  O="+Str(FO)
pString=@String$
SetWindowText_(hWnd, pString)
!POPAD                  ;die gesicherten Register wieder zurck holen
!POPFD                  ;die gesicherten Flags wieder zurck holen
EndMacro 
;--------
; Also einfach an der interssierenden Code-Stelle "asminfo" als neue Zeile einfgen.
; Funktioniert nicht bei zugeschaltetem Debugger (dafr gibt es CALLDEBUGGER).
; Die angezeigten Register-Werte sind in Hex!

;=========================================================================================

; SONSTIGE VORAUSSETZUNGEN
; Kenntnisse ber Zahlensysteme (Dezimal, Hexadezimal, Binr) sowie ber Bits, Bytes,
; Words, Double-Words usw. sollten vorhanden sein.
; Eine Hexadezimal-Zahl in FAsm wird durch ein nachgestelltes h (oder H) dargestellt; 123
; ist eine Dezimal-Zahl und 123h ist eine Hexadezimal-Zahl. 11001101b ist eine Binr-Zahl. 
; Fr Hexadezimal-Zahlen sind auch die Vorstze 0x und $ mglich; also 0x123 und $123.
; Mathematische Begriffe:
; - Addition: Summand + Summand = Summe
; - Subtraktion: Minuend - Subtrahend = Differenz
; - Multiplikation: Multiplikator x Multiplikand = Produkt
; - Division: Dividend / Divisor = Quotient
; Sonstige Begriffe:
; - MSB: Most Significant Bit; das hchstwertige Bit eines Registers oder Variable
; - LSB: Lowest Significant Bit; das niedrigstwertige Bit eines Registers oder Variable (also
;        immer Bit0)
; Die Bit-Zhlweise ist wie unten fr ein Byte dargestellt:
;   +----+----+----+----+----+----+----+----+
;   |Bit7|Bit6|Bit5|Bit4|Bit3|Bit2|Bit1|Bit0|   ein Double-Word z.B. geht von Bit0 bis Bit31
;   +----+----+----+----+----+----+----+----+
;    MSB                                 LSB
;    128   64   32   16    8    4    2    1     Zhl-Wert

; Der Zhl-Wert eines Bits betrgt: Wert = 2^Bitn; also z.B. fr Bit3 = 2^Bit3 = 2^3 = 8.
; Der Wert eines Bytes (Words, DWords) ist die Summe seiner Bit-Werte.  
; Ein Byte besteht aus 8, ein Word aus 16 und ein DWord (Double-Word) aus 32 Bits.
; Ein Byte kann man noch aufteilen in ein Low-Nibble (Bits 0-3) und ein High-Nibble (Bits 4-7). 

;=========================================================================================

; ALLGEMEINES
; Da der PB-Wortschatz an Assembler-Befehlen seit Jahren leider nicht weiter 
; aktualisiert wurde empfehle ich die Verwendung von "!" zur Einleitung einer
; Assembler-Befehlszeile. Dies sieht auch bei Verwendung nicht von PB untersttzter
; Befehle einheitlicher und bersichtlicher aus. Das "!" bewirkt eine direkte
; Befehlsbergabe an FAsm. Es muss dann auch nicht explizit die ASM-Untersttzung
; in den Optionen zugeschaltet sein.
; Generell sollte vermieden werden, Labels, Prozeduren, Arrays, Variablen usw. mit
; Assembler-Befehlswrtern zu benennen. Oft gesehenes (schlechtes) Beispiel dafr ist
; "TEST", was bei zugeschaltetem Inline-ASM z.B. als Prozedur-Name zu den tollsten
; Fehlermeldungen fhrt.
; Dies ist kein FAsm-Tutorial und nicht alle Feinheiten von FAsm sind in PB so ohne
; weiteres nutzbar (wenn berhaupt); aber das ist fr unsere Zwecke ohne Belang. PB ist
; schliesslich keine Entwicklungsumgebung fr FAsm!

;=========================================================================================

; LOS GEHTS
; bliche ASM-Tuts fangen mit Erklrungen zu Segmenten, Speichermodellen, DOS-Interrupts
; usw. an. Damit brauchen (und wollen) wir uns hier berhaupt nicht beschftigen, denn
; alles was wir zum Ausfhren unserer ASM-Codes brauchen, bringt PB schon mit. Und zwar
; so schn, dass wir die Ergebnisse gleich unter Windows betrachten knnen!

;=========================================================================================

; THEORIE
; Ein wenig Theorie muss sein. Deshalb gestrafft ein berblick ber Prozessor-Register und
; Variablen.

;=========================================================================================

; PROZESSOR-REGISTER
; 64-Bit-Prozessoren werden hier wie 32-Bit-Prozessoren betrachtet (Register!).
; Fr uns von Interesse sind hier die sogenannten General-(Standard-)Register und
; dazu noch das (wichtige!) (E)FLAG-Register. Das "E" steht fr Extended, Expanded, also
; "Erweitert" und bedeutet hier "grssenmssig (erweitert auf) 32-Bit". "(E)FLAG" oder bei
; Registern z.B. "(E)AX" bedeuten also "gltig fr 32- und 16-Bit". 
; Die General-Register sind die eigentlichen Arbeitssttten des Prozessors. Hier hinein
; werden Daten geladen, verarbeitet und stehen dann zum Auslesen bereit.
;-----------------------------------------------------------------------------------------
; GENERAL-REGISTER
; Die General-Register sind 32-Bit breit und haben folgende Namen:
; - EAX  auch Akkumulator-Register genannt, die unteren 16-Bit sind ansprechbar als AX
;        (Bits 0-15). AX lsst sich weiter aufteilen in die zwei 8-Bit-Register AL
;        (L=Low, Bits 0-7) und AH (H=High, Bits 8-15) 
; - EBX  auch Base-Register genannt, Aufteilungen wie bei EAX mglich (BX, BL und BH)
; - ECX  auch Count-Register genannt, Aufteilungen wie bei EAX mglich (CX, CL und CH)
; - EDX  auch Data-Register genannt, Aufteilungen wie bei EAX mglich (DX, DL und DH)
; Die obigen 4 Register werden auch als Universal-Register bezeichnet.
; - ESI  auch Source-Index genannt, die unteren 16-Bit sind ansprechbar als SI 
; - EDI  auch Destination-Index genannt, die unteren 16-Bit sind ansprechbar als DI 
; Die obigen 2 Register werden auch als Index-Register bezeichnet.
; - EBP  auch Base-Pointer-Register genannt, die unteren 16-Bit sind ansprechbar als BP
; - ESP  auch Stack-Pointer-Register genannt, die unteren 16-Bit sind ansprechbar als SP
; - EIP  auch Instruction-Pointer-Register genannt, die unteren 16-Bit sind ansprechbar als IP
; Die obigen 3 Register werden auch als Pointer-Register bezeichnet.
; Beispiel: Die 32-Bit-Universal-Register sind so aufgebaut:
;   +------------------------------------------------------------------+
;   |                          Bit31 - Bit0                            | = bei Verwendung als  
;   |                      (EAX, EBX, ECX, EDX)                        |   32-Bit-Register
;   +------------------------------------------------------------------+
;   +--------------------------------+---------------------------------+
;   |                                |           Bit15 - Bit0          | = bei Verwendung als
;   |                                |         (AX, BX, CX, DX)        |   16-Bit-Register
;   +--------------------------------+---------------------------------+
;   +--------------------------------+----------------+----------------+
;   |                                |  Bit15 - Bit8  |  Bit7 - Bit0   | = bei Verwendung als 
;   |                                |(AH, BH, CH, DH)|(AL, BL, CL, DL)|   8-Bit-Register
;   +--------------------------------+----------------+----------------+
; Bit31                         Bit16 Bit15       Bit8 Bit7           Bit0

; Wenn z.B. von AL die Rede ist sind also die unteren 8 Bit vom Register EAX (oder auch AX)
; gemeint und z.B. MOV AL,123 beeinflusst nur das Register AL, der Rest von (E)AX bleibt
; unverndert. Oben nicht bezeichnete Register-Teile (Bit16-31) knnen in den 16-bzw.8-Bit-
; Anwendungsfllen schlicht ignoriert werden oder vom Programmierer z.B. als Zwischenspeicher
; genutzt werden.   

; Die Universal-Register (nomen est omen) sind fr (fast) alle Zwecke zu gebrauchen. Sie
; sind neben ihrem Zweck als Daten-Register auch als Zeiger in den Speicher einsetzbar
; (Ausnahmen s.u.: Stack und Stringbefehle). 
; Das 8-Bit-Register CL (Bits 0-7 von ECX) hat fr einige Befehle eine gesonderte Funktion
; als Zhl-Register und fr ECX existieren auch einige Sonderflle. 

; Die Index-Register werden fr diverse String-Befehle als Adressierungs-Zeiger bentigt.
; ESI muss auf die Quelle (Source) und EDI auf das Ziel (Destination) zeigen. Ansonsten
; kann man diese Register aber auch wie die Universal-Register benutzen. 

; Der Stackpointer ist der Zeiger in den Stack (na so was!). Der Stack (auch Stapel- oder
; Keller-Speicher genannt) ist ein reservierter Speicherbereich, in dem z.B. Rcksprung-
; Adressen von Call-Aufrufen hineingeschrieben werden. Die Stack-Adressierung ist ausser
; mit dem Stackpointer auch mit dem Base-Pointer (EBP) mglich. EBP kann man hnlich wie
; die Universal-Register fr alles mgliche verwenden, als Adress-Zeiger zeigt es aber 
; immer auf den Stack! Weiteres siehe beim Beispiel fr PUSH/POP.

; Der Instruction-Pointer zeigt innerhalb des Programms auf den nchsten auszufhrenden
; Befehl und wir sollten tunlichst nicht versuchen ihn zu manipulieren. In PB gibt es z.B.
; die Befehle "GetCurrentEIP" und "GoToEIP", aber wie gesagt, soll hier nicht weiter behandelt
; werden. 
;-----------------------------------------------------------------------------------------
; FPU-REGISTER
; FPU steht fr Floating-Point-Unit, also fr die Fliesskomma-Zahlen-Einheit. Die FPU
; enthlt 8 eigene, 80-Bit breite Register, die mit ST0 bis ST7 bezeichnet werden, wobei
; STx nur "Platzhalter" sind, also nicht direkt die Register angesprochen werden. Die
; Verwaltung bernimmt die FPU selbst und hat uns hier nicht weiter zu interessieren.  
;-----------------------------------------------------------------------------------------
; MMX-REGISTER
; MMX steht fr Multi-Media-Extension und sind physikalisch die selben Register wie bei
; der FPU, von denen aber nur 64-Bit genutzt werden und MMX0 bis MMX7 benannt sind, wobei
; hier direkt die Register angesprochen werden.
;-----------------------------------------------------------------------------------------
; XMM-REGISTER
; XMM steht fr Extended-Multi-Media und es sind 8 echt neue eigenstndige 128-Bit-Register
; fr die SSE-Befehle (SSE=Streaming SIMD Extensions; SIMD=Single Instruction, Multiple Data,
; d.h. mit einem Befehl knnen mehrere Daten gleichzeitig bearbeitet werden).
;-----------------------------------------------------------------------------------------
; EFLAG-REGISTER
; Dies ist ein 32-Bit breites Register, dessen einzelne Bits z.B. das Ergebnis von
; gerade durchgefhrten mathematischen Operationen widerspiegeln. Nicht alle Bits dieses
; Registers werden genutzt, aber die verwendeten sollte man weitestgehend kennen, da sie
; wichtig fr bedingte Sprnge, Vorzeichen-Erkennung, ber-/Unter-Lufe usw. sind.
; Gesetzt werden diese Flags (=Bits) nur durch Operationen mit den General-Registern oder
; spezielle Befehle (Ausnahmen gibt es auch hier, sind hier aber nicht von Bedeutung).
; Wegen ihrer Wichtigkeit hier die Grund-Flags:
; - Bit0  = Carry-Flag (C), zeigt z.B. ber- oder Unterlufe ("Borgen" bei der Subtraktion) 
;                           an; bei Bit-Schiebe-Operationen zeigt dieses Flag den Wert des
;                           gerade herausgeschobenen Bits an  
; - Bit2  = Parity-Flag (P), zeigt an, ob die unteren 8 Bits eines Registers/Variable nach
;                            einer Operation eine gerade oder ungerade Anzahl gesetzter Bits
;                            haben   
; - Bit4  = Auxiliary-Flag (A), zeigt an, ob bei einer mathematischen Operation in einem Register
;                               oder Variablen ein berlauf von Bit3 zu Bit4 stattgefunden hat.
;                               Wichtig fr BCD-Operationen (BCD=Binary Coded Decimals); hier
;                               erstmal nicht weiter von Interesse
; - Bit6  = Zero-Flag (Z), zeigt an, ob das Ergebnis der Operation Null war
; - Bit7  = Sign-Flag (S), zeigt an, ob nach der Operation das hchstwertige Bit des Ergebnisses
;                          gesetzt ist. Diese Bit zeigt bei vorzeichenbehafteten Werten im 
;                          gesetzten Fall an, das der Wert negativ ist  
; - Bit8  = Trap-Flag (T), kann ber Umwege vom Nutzer gesetzt werden, den Prozessor in den
;                          Single-Step-Modus zu versetzen, um das jeweilige Programm schrittweise
;                          abzuarbeiten (Fehlersuche, Debugger)
; - Bit9  = Interrupt-Enable-Flag (I), wenn gelscht, ignoriert der Prozessor anstehende Interrupt-
;                                      Anforderungen (von z.B. Hardware) und fhrt unbeirrt im
;                                      Programm-Ablauf fort. Sowas lsst sich natrlich aushebeln...  
;                                      Ist fr uns hier auch nicht weiter von Interesse
; - Bit10 = Direction-Flag (D), fr String-Befehle sehr wichtiger Befehl; hiermit wird die Lese- 
;                               und Schreib-Richtung festgelegt. Ist das Flag gelscht, wird nach
;                               jeder String-Operation die Lese-und-Schreib-Adresse erhht, sonst
;                               verringert      
; - Bit11 = Overflow-Flag (O), wie der Name schon sagt, werden berlufe angezeigt, z.B. wenn das
;                              Ergebnis einer Multiplikation (mit den General-Registern!) nicht 
;                              mehr in das Ergebnis-Register passt. Es wird auch gesetzt, wenn
;                              ein berlauf in das MSB (Vorzeichen-Bit!) stattgefunden hat 

;=========================================================================================

; DEKLARATION VON VARIABLEN
; Assembler-Variablen mssen vor Verwendung deklariert werden! In reinen FAsm-Programmen gibt
; es dafr die "section '.data' Data readable writeable", in der alle verwendeten Variablen
; mit ihrer Grsse und eventuellen Vorwerten aufgefhrt sind. Auf so deklarierte Variablen kann
; PB aber nicht zugreifen; sie existieren fr PB berhaupt nicht. Ein Ausweg wre, vor jeder
; Variable einen Label-Namen zu setzen, den PB erkennen kann. Aber so kompliziert wollen wir es
; vorerst nicht machen; wir deklarieren einfach die bentigten Variablen in PB als "Global" und
; PB und auch FAsm knnen direkt darauf zugreifen. Die Assembler-Variable muss dann in der
; Form "[v_Variable]" geschrieben werden (denn so legt PB sie an), aber das ist reine bungssache.        
; WICHTIG: In FAsm sind Variablen-Namen Case-Sensitiv; d.h. es wird streng zwischen Gro- und
; Klein-Schreibung unterschieden: "Var1" ist in FAsm eine andere Variable als "VAr1" !

;=========================================================================================

; PRAXIS
;=========================================================================================
 
; Das berhmt-berchtigte "Hello World!" als Einstieg lassen wir mal weg, das gehrt zur
; Beschreibung einer nativen Programmiersprache.

; DIE WICHTIGSTEN ASSEMBLER-BEFEHLE
;=========================================================================================

; Da fangen wir mal mit dem MOV-Befehl an. 
; Syntax: MOV Ziel,Quelle. Beeinflusste Flags: Keine
; Mit MOV wird der Wert, den Quelle enthlt, in Ziel hineingeschrieben. Der Wert von Quelle
; wird nicht verndert.
; Beispiel: (die hier deklarierten Variablen gelten auch fr alle anderen Beispiele!)
;--------
Global Var1.l=20        ;Typ l=Long entspricht in FAsm DWord, also Double-Word = 32-Bit
Global Var2.l=10 
!MOV eax,[v_Var1]       ;der Wert von Variable Var1 wird in das Prozessor-Register EAX geladen
!MOV [v_Var2],eax       ;jetzigen Wert von EAX in Variable Var2 schreiben
MessageRequester("MOV", "Var2 hatte den Wert 10 und jetzt: " + Str(Var2))
;--------
; Zur Bildschirmausgabe benutzen wir ganz bequem den MessageRequester von PB. 
; Als Operanden knnen auch nur Register verwendet werden:
;--------
!MOV ebx,edx            ;beliebige Register, die wenigen Einschrnkungen s.o.(z.B. EIP)
!MOV [v_Var1],ebx       ;EBX hat soeben den Wert von EDX erhalten und dieser Wert wird 
                        ;in Var1 geschrieben  
MessageRequester("MOV", "Das Register EDX hatte den zuflligen Wert: " + StrU(Var1,#PB_Long))
;--------
; Nicht mglich ist z.B. "MOV [v_Var1],[v_Var2]".
; Das letzte Beispiel kann man natrlich auch einfacher und somit schneller schreiben:
;--------
!MOV [v_Var1],edx
MessageRequester("MOV", "Das Register EDX hatte jetzt den zuflligen Wert: " + StrU(Var1,#PB_Long))
;--------
; Und was sieht man: Obwohl wir den Wert vom Register EDX nicht verndert haben (die Quelle
; bleibt ja unverndert), zeigen die letzten beide Beispiele fr EDX unterschiedliche Werte
; an. "Schuld" daran sind eingeschobene PB-Befehle (hier Messagerequester und StrU), die
; ja auch nichts anderes machen als Prozessor-Register zu benutzen. Das ist beim Ablauf von
; Assembler-Code-Teilen immer zu beachten! Fr die Sicherung noch weiterhin bentigter
; Register-Werte ist der Programmierer selbst verantwortlich. Man knnte nun vorher den 
; noch bentigten Register-Wert in eine weitere Variable schreiben (sichern) und hinterher
; wieder zurckholen, aber es geht hier einfacher mit den PUSH-und POP-Befehlen:
; Syntax: PUSH Quelle. Beeinflusste Flags: Keine
;         POP Ziel. Beeinflusste Flags: Keine 
; Quelle und Ziel knnen Register oder Variablen sein, fr PUSH auch Zahlenwerte.
; PUSH Quelle schreibt den Wert auf den Stack (adressiert mit dem Stack-Pointer ESP). Da
; der Stack von oben nach unten (Speicher-Adressen-Mssig gesehen) beschrieben wird, wird
; mit einem PUSH (32-Bit-)Quelle der ESP um 4 verringert und zeigt auf die nchste Schreib-
; Adresse. Der Stack ist als FILO-Speicher organisiert, also First In, Last Out. Der zuerst
; reingeschriebene Wert wird als letzter wieder rausgeholt.
; POP Ziel erhht den Stackpointer ESP wieder um 4 (bei 32-Bit-Werten; dabei bleiben wir mal)    
; und liest den mittels ESP-adressierten Wert wieder vom Stack aus und schreibt ihn in Ziel.
; Kennen sollte man noch die "Sammelbefehle" PUSHAD und POPAD, die bis auf EIP alle anderen
; 8 General-Register in einem Rutsch auf den Stack schreiben bzw. lesen. 
; PUSHA (=PUSHAW) und POPA (=POPAW) sind fr PUSHAD bzw. POPAD die Entsprechungen fr 16-Bit-
; Register und interessieren uns nicht weiter. Gleiches gilt fr die PUSH-/POP-Befehle, die
; das (E-)Flag-Register betreffen (PUSHF, PUSHFD, PUSHFW bzw. POPF, POPFD und POPFW). 
; So, jetzt wollen wir das mal anwenden:
;--------
!MOV [v_Var1],edx       ;das kennen wir schon
!PUSH edx               ;der Wert von Register EDX wird auf den Stack geschoben 
MessageRequester("PUSH/POP", "Das Register EDX hatte den zuflligen Wert: " + StrU(Var1,#PB_Long))
!POP edx                ;der mit PUSH gesicherte Wert wird wieder zurckgelesen  
!MOV [v_Var1],edx       ;fr die Anzeige wieder in Var1 schreiben; mit dem alten Wert von EDX
                        ;knnte jetzt weiter gearbeitet werden
MessageRequester("PUSH/POP", "Mit POP zurckgeholter Wert von Register EDX: " + StrU(Var1,#PB_Long))       
;--------
; Anmerkung: Die PB-Befehle verwenden natrlich auch fleissig PUSH/POP und htten Fred&Co
; hier Mist gebaut, wrde unser ausgelesener Wert nicht stimmen. Die Anzahl von PUSH und POP
; muss (am Ende des Programms oder Programm-Teiles) gleich sein, sonst greifen nachfolgende
; Code-Teile auf falsche Werte zurck! Dafr ist der Programmierer selbst verantwortlich.  
; Wichtig ist, sich ber die Reihenfolge im klaren zu sein (s.o., in Umkehrung: Der zuletzt 
; reingeschriebene Wert wird als erster wieder ausgelesen). 
; Man darf bei PUSH/POP nicht den berblick verlieren, also wer unsicher ist soll zum Sichern
; doch besser extra Variablen verwenden. Auf dem Stack landen ja nicht nur Push-Werte, sondern
; auch z.B.die Rcksprung-Adressen von Call-Aufrufen (oder GOSUB und Prozedur-Aufrufe in PB)
; oder an aufgerufene Prozeduren bergebene Parameter.

; Zu den MOV-Befehlen gehren auch die String-Mov-Befehle MOVS, MOVSB, MOVSW und MOVSD, mit
; denen mittels der Zeiger ESI und EDI eben Strings umkopiert werden knnen. Sind aber erstmal
; uninteressant. Von grsserer Bedeutung sind MOVSX und MOVZX, mit denen Operanden-Erweiterungen
; durchgefhrt werden knnen:
; Syntax: MOVSX Operand1, Operand2. Flags werden nicht verndert.
; Operand1 muss ein 16-oder 32-Bit-Register sein, Operand2 kann ein 8-oder 16-Bit-Register
; oder Variable sein.
; Operand1 muss grsser sein als Operand2, also wenn Operand2 Word-Gre hat, muss Operand1 
; DWord-Grsse haben. Die Erweiterung erfolgt vorzeichenbehaftet; also in den "Erweiterungsteil"
; von Operand1 wird das Vorzeichen-Bit von Operand2 geschrieben:
;--------
!MOV ax,-12345
!MOVSX ebx,ax
!MOV [v_Var1],ebx
MessageRequester("MOVSX", "Das Bitmuster sieht so aus : " + RSet(Bin(Var1,#PB_Long),32,"0"))
;--------
; Oben zum Test mal das Vorzeichen von AX ndern!

; Syntax: MOVZX Operand1, Operand2. Flags werden nicht verndert.
; Operand1 muss ein 16-oder 32-Bit-Register sein, Operand2 kann ein 8-oder 16-Bit-Register
; oder Variable sein.
; Operand1 muss grsser sein als Operand2, also wenn Operand2 Word-Gre hat, muss Operand1 
; DWord-Grsse haben. Die Erweiterung erfolgt vorzeichenlos; also in den "Erweiterungsteil"
; von Operand1 werden Nullen geschrieben:
;--------
!MOV ax,-12345
!MOVZX ebx,ax
!MOV [v_Var1],ebx
MessageRequester("MOVZX", "Das Bitmuster sieht so aus : " + RSet(Bin(Var1,#PB_Long),32,"0"))
;--------
; Auch hier mal das Vorzeichen von AX ndern!

; So, jetzt gehts flott weiter: Daten hin-und-herschieben befriedigt auf Dauer nicht. Wir
; wollen Daten verndern, Berechnungen durchfhren usw. Aber: Ganz wichtig sind die Sprung-
; Befehle:
; Syntax: JMP Sprungziel. Beeinflusste Flags: Keine
; JMP ist ein sogenannter unbedingter Sprung, d.h. der Sprung wird auf jeden Fall ausgefhrt;
; er ist von keinerlei Bedingungen abhngig:
;--------
!MOV [v_Var1],50
!JMP l_label1           ;springt ohne Wenn und Aber zu Label1
!MOV [v_Var1],500       ;diese Programmzeile wird nicht abgearbeitet!
Label1:                 ;Sprungziel
MessageRequester("JMP mit PB-Konvention", "Variable Var1 hat den Wert: " + StrU(Var1,#PB_Long)) 
;--------
; Anmerkung: Als Label fr das Sprungziel verwenden wir obiges Verfahren, so dass  PB als
; auch FAsm darauf zugreifen knnen (dafr sorgt der PB-Compiler). Dafr ist aber notwendig,
; in Assembler das Sprungziel als "Sprung l_label" zu schreiben; und zwar muss der Label-Name
; im Sprungbefehl unbedingt klein geschrieben werden!
; Wenn das Sprungziel nur vom Assemble-Teil genutzt wird kann man auch schreiben: 
;--------
!MOV [v_Var1],60
!JMP Label2             ;springt ohne Wenn und Aber zu Label2
!MOV [v_Var1],600       ;diese Programmzeile wird nicht abgearbeitet!
!Label2:                ;Sprungziel
MessageRequester("JMP mit Nur-FAsm-Konvention", "Variable Var1 hat den Wert: " + StrU(Var1,#PB_Long)) 
;--------
; Wichtig: Der Name eines Sprungzieles muss in einem Programm eindeutig sein (hier: Label1
; und Label2). Bei Doppel-Belegung gibts zwar Mecker, aber man muss nach einer Label-
; Namensnderung auch daran denken, berall wo erforderlich den neuen Label-Namen einzutragen.
; Klingt vielleicht doof, ist aber ein gern gemachter Flchtigkeits-Fehler.
; Bevor jemand meckert: Im letzten Beispiel knnte auch wieder "Label1" verwendet werden, weil
; eben FAsm spezifisch. Trotzdem sollte man es nicht machen! 
; Zu den unbedingten Sprngen kann man auch CALL und RET zhlen. Mit CALL wird zu einer Sub-
; (Unter-) Routine gesprungen und mit RET erfolgt aus dieser Sub-Routine der Rcksprung zum
; Befehl nach CALL. Dies entspricht in PB den Befehlen GOSUB und RETURN. Genau wie GOSUB sollte
; man CALL nicht in Prozeduren verwenden!

; Wenn es unbedingte Sprnge gibt, existieren auch bedingte Sprnge. Derer gibt es eine ganze 
; Menge und als Sprung-Kriterien (also ob gesprungen wird oder nicht) fungieren hauptschlich
; die oben schon erluterten Flags. Und da deren Zustand hauptschlich von den mathematischen
; Operationen der (General-)Register oder Variablen abhngt, verbinden wir beides.
; Grundstzlich gilt: Alle hier aufgefhrten Operationen sind Integer-Operationen! Fliekomma-
; Operationen sind nur mit den FPU-oder SSE-Registern mglich.
;--------
!MOV eax,99             ;klar
!MOV [v_Var1],99        ;klar
!ADD [v_Var1],eax       ;neu: Zum Wert von Var1 wird der Wert von EAX zuaddiert
!JZ l_label3            ;neu: Springe zu Label3, wenn das Ergebnis Null ist
MessageRequester("Addition und bedingter Sprung", "Das Ergebnis lautet: " + StrU(Var1,#PB_Long)) 
Label3:                 ;hier knnte man melden: Ergebnis war Null

; Syntax: ADD Ziel,Quelle. Beeinflusste Flags: C, P, A, Z, S, O
; ADD sollte klar sein: Zum Wert von Ziel wird der Wert von Quelle zuaddiert. Quelle bleibt
; unverndert und nicht beide Operanden drfen Variablen sein. Die Werte sind Integer-Werte,
; zu den Vorzeichen kommen wir noch. Quelle kann auch ein direkter Wert (Konstante) sein.
; Syntax: JZ Sprungziel
; JZ bedeutet "Springe, wenn das Zero-(Null-)Flag gesetzt ist", also das Ergebnis der letzten
; Operation den Wert Null hatte (denn dann wird das Flag gesetzt). War das Ergebnis nicht
; Null, wird das Flag gelscht (und nicht etwa unberhrt gelassen).
; Da das Ergebnis grsser als Null war, wird der (bedingte) Sprung nicht ausgefhrt und wir
; knnen uns an der Ergebnis-Ausgabe ergtzen.
; Um den Sprung mittels JZ auszufhren knnte man einen der obigen Operanden einfach negativ
; machen, aber wir wollen es mit der Subtraktion testen:
;--------
!MOV eax,99             ;klar
!MOV [v_Var1],99        ;klar
!SUB [v_Var1],eax       ;neu: Vom Wert von Var1 wird der Wert von EAX subtrahiert
!JZ l_label4            ;springe zu Label4, wenn das Ergebnis Null ist
MessageRequester("Subtraktion und bedingter Sprung, wird nicht angezeigt", "Das Ergebnis lautet: " + StrU(Var1,#PB_Long)) 
!JMP l_label5           ;um die nchste (dann falsche) Meldung nicht anzuzeigen
Label4:
MessageRequester("Subtraktion und bedingter Sprung, wird angezeigt", "Das Ergebnis lautet: " + StrU(Var1,#PB_Long))
Label5:
;--------
; Syntax: SUB Ziel,Quelle. Beeinflusste Flags: C, P, A, Z, S, O
; Der Wert von Quelle wird von Ziel subtrahiert, Quelle bleibt unverndert. Sonst wie ADD.

; Der JZ-Befehl kommt aber noch bei einem ganz wichtigen Befehl zum tragen; dem CMP-Befehl
; (CMP=compare, also vergleiche). Der CMP-Befehl ist nmlich fr den Prozessor nichts anderes
; als eine Subtraktion (wie oben), bei der aber beide Operanden unverndert bleiben.
; Entsprechend dem Ergebnis werden aber die entsprechenden Flags gesetzt.
; Syntax: CMP Ziel,Quelle. Beeinflusste Flags: C, P, A, Z, S, O
; Quelle wird von Ziel subtrahiert, beide Operanden bleiben aber unverndert. Sonst wie ADD.
;--------
!MOV eax,99
!MOV [v_Var1],99
!CMP [v_Var1],eax       ;neu: Vergleiche Wert von Var1 mit Wert von EAX
!JZ l_label6            ;springe zu Label6, wenn beide Werte gleich sind
MessageRequester("CMP und Test auf Gleichheit", "Die Werte sind nicht gleich!") 
!JMP l_label7           ;um die nchste (dann falsche) Meldung nicht anzuzeigen
Label6:
MessageRequester("CMP und Test auf Gleichheit", "Die Werte sind gleich!")
Label7: 
;--------
; In FAsm knnen auch anonyme Label verwendet werden. Diese werden mit dem "@"-Zeichen 
; gebildet und sind z.B. in oft genutzten Prozeduren von Nutzen, um Namens-Kollisionen mit
; dem Rest-Code zu vermeiden. Gesprungen werden kann vorwrts mit @f (f=forward) und zurck
; mit @b (b=backward) oder @r (r=reverse), aber immer nur bis zum ERSTEN gefundenen Label @@.
; Als Beispiel hier eine Schleife mit Register ECX als Schleifen-Zhler:
;--------
!MOV ecx,5
!@@:                     ;anonymes Label1
;irgendwas ausfhren
!CMP eax,12345           
!JE @f                   ;springt vorwrts zum anonymen Label2, wenn EAX den Wert 12345 hat (vorzeitiges Verlassen der Schleife)
;irgendwas ausfhren
!SUB ecx,1
!JNZ @b                  ;springt zurck zum anonymen Label1, solange ECX nicht Null ist
!@@:                     ;anonymes Label2
;irgendwas
!@@:                     ;anonymes Label3, ist fr den obigen Code nicht erreichbar! 
;--------
; Mitunter kommt man mit anonymen Labeln nicht weit, weil die Programm-Struktur doch komplizierter ist.
; In Proceduren und Macros ist dann die Verwendung der FAsm-Directive "local" angesagt:
Macro Irgendwas(Wert_X, Wert_Y)        ;oder sonstwas
  !local X_Loop                        ;damit kann man dieses Macro an mehreren Programm-Stellen einsetzen          
  !local Y_Loop                        ; ohne das es Mecker wegen bereits vorhandenen Labels gibt
  ;irgendwas
!X_Loop:
  ;irgendwas; wo z.B. die anonymen Label schon ausgereizt sind 
!Y_Loop:
  ;irgendwas
EndMacro
;--------

; Wenn auf Gleichheit getestet werden soll, wird oft auch der Befehl JE (jump if equal)
; benutzt. Wenn beide Werte gleich sind, ist ja das Subtraktions-Ergebnis Null. 
; Hier ist wichtig zu wissen, das die Assembler-Befehle JZ (springe wenn Null) und
; JE (springe wenn gleich) fr den Prozessor vllig identisch sind! Man kann JZ durch JE
; ersetzen, wenn es dem besseren Verstndnis im konkreten Fall dient.

; Zu den Vergleichs-Befehlen gehrt auch CMPXCHG, an dem 3 Operanden beteiligt sind, wobei
; einer davon implizit das AL-, AX- oder EAX-Register ist, je nach Grsse der beiden anderen
; Operanden. Wie aus dem Namem schon ersichtlich ist, wird nicht einfach nur verglichen, sondern
; auch getauscht:
; Syntax: CMPXCHG Operand1,Operand2. Beeinflusste Flags: C, P, A, Z, S, O
; Operand 2 muss ein Register sein, Operand1 kann ein Register oder Variable sein, Beide
; Operanden mssen vom selben Grssen-Typ sein und bestimmen so auch, welcher Teil des Akkumulators
; verwendet wird. Verglichen wird Operand1 mit dem Akkumulator, bei Gleichheit wird Operand2 
; in Operand1 kopiert; der Akkumulator bleibt unverndert. Bei Ungleichheit wird Operand1 in
; den Akkumulator kopiert, Operand2 bleibt unverndert:
;--------
!MOV eax,1234           ;EAX als impliziter Operand 
!MOV ebx,1234           ;Operand1
!MOV ecx,5678           ;Operand2
!CMPXCHG ebx,ecx        ;vergleicht EAX mit EBX (Operand1)
!MOV [v_Var1],ebx
MessageRequester("CMPXCHG", "EAX und Operand1 sind gleich, deshalb erhlt Operand1 den Wert von Operand2 : " + Str(Var1))
;--------
;--------
!MOV eax,12345          ;EAX als impliziter Operand 
!MOV ebx,1234           ;Operand1
!MOV ecx,5678           ;Operand2
!CMPXCHG ebx,ecx        ;vergleicht EAX mit EBX (Operand1)
!MOV [v_Var1],eax
MessageRequester("CMPXCHG", "EAX und Operand1 sind nicht gleich, deshalb erhlt EAX den Wert von Operand1 : " + Str(Var1))
;--------
; CMPXCHG gibt es auch ein einer Version, die 8 Bytes austauscht, aber das schenken wir uns hier.

; Mit dem CMP-Befehl knnen wir auch gleich einen weiteren bedingten Sprung-Befehl testen:
;--------
!MOV eax,99
!MOV [v_Var1],100
!CMP [v_Var1],eax       ;vergleiche Wert von Var1 mit Wert von EAX
!JA l_label8            ;neu: springe zu Label8, wenn Var1 grsser ist als Wert von EAX
MessageRequester("CMP und Test auf grsser", "Var1 ist nicht grsser als Wert von EAX!") 
!JMP l_label9           ;um die nchste (dann falsche) Meldung nicht anzuzeigen
Label8:
MessageRequester("CMP und Test auf grsser", "Var1 ist grsser als Wert von EAX!")
Label9: 
;--------
; Syntax: JA Sprungziel
; Springe, wenn weder das Zero- noch das Carry-Flag gesetzt sind nach der letzten Operation,
; hier also dem CMP. Zero-Flag ist klar, das Carry-Flag wird bei Subtraktionen gesetzt, wenn
; das Ergebnis negativ ist ("Borgen"). Also: Springe, wenn das Ergebnis der (im Falle von 
; CMP fiktiven) Subtraktion weder Null noch negativ ist. Da JA nicht das Sign-Flag (Vorzeichen)
; auswertet folgt auch, dass JA nur fr vorzeichenlose Integer-Werte zu benutzen ist.     

; Jetzt alle mglichen Varianten mit Beispielen zu belegen wrde den Rahmen sprengen, deshalb
; hier folgende bersicht:

; JA      jump if above                springe wenn grsser, vorzeichenlos 
; JAE       "  "  above or equal          "     "   grsser oder gleich, vorzeichenlos
; JB        "  "  below                   "     "   kleiner, vorzeichenlos
; JBE       "  "  below or equal          "     "   kleiner oder gleich, vorzeichenlos
; JC        "  "  carry                   "     "   Carry-Flag gesetzt
; JE        "  "  equal                   "     "   gleich
; JG        "  "  greater                 "     "   grsser, vorzeichenbehaftet
; JGE       "  "  greater or equal        "     "   grsser oder gleich, vorzeichenbehaftet
; JL        "  "  less                    "     "   kleiner, vorzeichenbehaftet
; JLE       "  "  less or equal           "     "   kleiner oder gleich, vorzeichenbehaftet
; JNA       "  "  not above               "     "   nicht grsser, vorzeichenlos
; JNAE      "  "  not above or equal      "     "   nicht grsser oder gleich, vorzeichenlos 
; JNB       "  "  not below               "     "   nicht kleiner, vorzeichenlos
; JNBE      "  "  not below or equal      "     "   nicht kleiner oder gleich, vorzeichenlos
; JNC       "  "  not carry               "     "   Carry-Flag nicht gesetzt
; JNE       "  "  not equal               "     "   nicht gleich
; JNG       "  "  not greater             "     "   nicht grsser, vorzeichenbehaftet
; JNGE      "  "  not greater or equal    "     "   nicht grsser oder gleich, vorzeichenbehaftet
; JNL       "  "  not less                "     "   nicht kleiner, vorzeichenbehaftet
; JNLE      "  "  not less or equal       "     "   nicht kleiner oder gleich, vorzeichenbehaftet
; JNO       "  "  not overflow            "     "   Overflow-Flag nicht gesetzt
; JNP       "  "  not parity              "     "   Parity-Flag nicht gesetzt
; JNS       "  "  not signum              "     "   Signum-Flag nicht gesetzt
; JNZ       "  "  not zero                "     "   Zero-Flag nicht gesetzt
; JO        "  "  overflow                "     "   Overflow-Flag gesetzt
; JP        "  "  parity                  "     "   Parity-Flag gesetzt
; JPE       "  "  parity even             "     "   Parity-Flag gesetzt
; JPO       "  "  parity odd              "     "   Parity-Flag nicht gesetzt
; JS        "  "  signum                  "     "   Signum-Flag gesetzt
; JZ        "  "  zero                    "     "   Zero-Flag gesetzt          

; Sprungbefehle, die in einer Zeile stehen, sind fr den Prozessor identisch und dienen nur der
; besseren Verstndlichkeit des Programm-Codes. 

; Sprungbefehl  |  ausgewertete Flags, springe wenn...                     
; -----------------------------------------------------------
; JA, JNBE      |  Carry und Zero beide nicht gesetzt
; JAE, JNB      |  Carry nicht gesetzt
; JB, JNAE, JC  |  Carry gesetzt
; JBE, JNA      |  Carry oder Zero gesetzt
; JE, JZ        |  Zero gesetzt
; JG, JNLE      |  Signum und Zero beide nicht gesetzt
; JGE, JNL      |  Signum nicht gesetzt
; JL, JNGE      |  Signum und Overflow ungleich
; JLE, JNG      |  Signum und Overflow ungleich oder Zero gesetzt
; JNC           |  Carry nicht gesetzt
; JNE, JNZ      |  Zero nicht gesetzt
; JNO           |  Overflow nicht gesetzt
; JNP, JPO      |  Parity nicht gesetzt
; JNS           |  Signum nicht gesetzt
; JO            |  Overflow gesetzt
; JP, JPE       |  Parity gesetzt
; JS            |  Signum gesetzt

; Fr alle Sprnge (bedingt und unbedingt) gilt: Kein Flag wird verndert!

; Zur Verdeutlichung von Vorzeichen (hchstes Bit eines Wertes): Ob die Register-oder 
; Variablen-Werte vorzeichenbehaftet oder nicht sind liegt ganz allein in der Interpretation
; des Programmierers! Der Prozessor kann nicht wissen was gemeint ist und setzt bei jeder
; Operation die Flags fr alle Flle. Die (richtige!) Auswertung liegt in der Hand des
; Programmierers!

; Auf zwei weitere bedingte Sprnge sei nur kurz hingewiesen: JCXZ und JECXZ. Hier werden
; keine Flags, sondern das CX-bzw.ECX-Register ausgewertet. Gesprungen wird, wenn das CX-
; bzw. ECX-Register den Wert Null hat. Im Gegensatz zu den o.g. Sprngen ist fr JCXZ und
; JECXZ der Sprungbereich stark eingeschrnkt: Das Sprungziel kann maximal 128 Bytes zurck
; oder 127 vorwrts sein. Gezhlt wird nach dem Sprung-Befehl, also zurck sind es noch
; weniger Bytes (Speicher-Adressen)! Erhlt man beim Ausfhren von Assembler-Code die Fehler-
; Meldung "error: code cannot be generated.", kann es durchaus ein zu grosser Sprung-Versuch
; gewesen sein. 

; Dann gibt es noch den Befehl LOOP mit seinen Ablegern LOOPE (=LOOPZ) und LOOPNE (=LOOPNZ).  
; Der Sprung wird ausgefhrt, wenn das Register ECX nicht Null ist; bei den Ablegern wird
; zustzlich noch das Zero-Flag ausgewertet. LOOP dekrementiert selbst das ECX-Register, ohne
; Flags zu verndern! Diese Befehle haben aber einen schlechten Ruf als Taktzyklen-Fresser
; und sollten durch andere Mglichkeiten ersetzt werden (einen eigenen Zhler auswerten und
; entsprechend springen). Auch hier gilt: Max. 128 Bytes zurck oder 127 Bytes vorwrts! 
; Soll doch der LOOP-Befehl eingesetzt werden und als Zhler das Register CX und nicht ECX
; verwendet werden, kann dies FAsm mit einem angehngtem "w" (fr Word) mitgeteilt werden;
; z.B LOOPW oder LOOPEW. Dadurch verringert sich aber der Sprungbereich zurck um 1! 

; So, jetzt gehts weiter mit den mathematischen Befehlen. ADD und SUB hatten wir schon,
; aber fr beide gibt es noch eine Variante, die einen aufgetretenen ber-bzw-Unterlauf 
; bercksichtigt. Fr die Addition gibt es ADC (Addition with Carry):
; Syntax: ADC Ziel,Quelle. Beeinflusste Flags: C, P, A, Z, S, O
;--------
Global Var3.q           ;wir benutzen hier auch eine Quad-Variable
; Hex-Werte zur besseren Darstellung!
!MOV edx,0h
!MOV eax,0ffffffffh
!MOV ebx,2h
!ADD ebx,eax            ;hier provozieren wir eine berlauf, das Ergebnis passt nicht mehr in
                        ; Register EBX und somit wird das Carry-Flag gesetzt
!ADC edx,0h             ;EDX wurde oben auf Null gesetzt, nun wird Null hinzuaddiert und auch
                        ; der Wert des Carry-Flags (0 oder 1) wird hinzuaddiert
!MOV dword[v_Var3],ebx  ;die Quad-Variable Var3 mit den 2 DWord-Teilen fllen
!MOV dword[v_Var3+4],edx
MessageRequester("ADC", Hex(Var3)) 
;--------
; Ruhig mal mit auskommentieren von ADC und/oder anderen Werten ausprobieren!
; Natrlich knnte man auch "per Hand" den Zustand des Carry-Flags nach der Addition ermitteln
; und verarbeiten, aber so gehts doch viel einfacher; ganz abgesehen davon, da in diesem Fall
; vorher das Carry-Flag in den gelschten Zustand gebracht werden msste. 

; Zu den Additions-Befehlen gehrt noch XADD:
; Syntax: XADD Operand1,Operand2. Beeinflusste Flags: C, P, A, Z, S, O
; Operand1 kann ein Register oder Variable sein, Operand2 muss ein Register sein. XADD vertauscht
; zuerst die Inhalte der Operanden, addiert sie dann und speichert das Ergebnis im 1.Operanden:
;--------
!MOV [v_Var1],67890
!MOV eax,12345
!XADD [v_Var1],eax
!MOV [v_Var2],eax
MessageRequester("XADD", "Das Ergebnis lautet " + Str(Var1) + " und EAX hat jetzt den Wert " + Str(Var2)) 
;--------

; Fr die Subtraktion mit Bercksichtigung eines eventuellen Unterlaufes gibt es den Befehl
; SBB (Subtraction With Borrow):
; Syntax: SBB Ziel,Quelle. Beeinflusste Flags: C, P, A, Z, S, O
; Als Beispiel hier eine Quasi-Quad-Subtraktion mit Long-Werten; der Minuend wird durch
; EDX:EAX und der Subtrahend durch ECX:EBX gebildet.
;--------
!MOV edx,32345678h   ;Minuend-High
!MOV eax,48765432h   ;Minuend-Low
!MOV ecx,23456789h   ;Subtrahend-High
!MOV ebx,56789abch   ;Subtrahend-Low
!SUB EAX,EBX         ;Minuend-Low - Subtrahend-Low
!SBB EDX,ECX         ;Minuend-High - Subtrahend-High unter Bercksichtigung eines evtl.
                     ; vorangegangenen Unterlaufs 
!MOV dword[v_Var3],eax  ;die Quad-Variable Var3 mit den 2 DWord-Teilen fllen
!MOV dword[v_Var3+4],edx
MessageRequester("SBB", Hex(Var3))
;--------
; Zum Ausprobieren SBB durch SUB ersetzen!
 
; Nach ADD und SUB mssen unweigerlich folgende zwei Befehle kommen: INC und DEC. INC erhht
; (inkrementiert) einen Wert um Eins und DEC verringert (dekrementiert) einen Wert um Eins.
; Syntax: DEC Operand. Beeinflusste Flags: P, A, Z, S, O  
; Operand kann ein Register oder Variable sein (jeweils 8-, 16- oder 32-Bit).
; Syntax: INC Operand. Beeinflusste Flags: P, A, Z, S, O  
; Operand kann ein Register oder Variable sein (jeweils 8-, 16- oder 32-Bit)
; Hier ist zu beachten: Im Gegensatz zu ADD oder SUB wird das Carry-Flag nicht verndert. Ist
; wichtig fr bedingte Sprnge (s.o.)!
; Wenn das Carry-Flag unwichtig ist sind diese Befehle schneller als ADD XXX,1 oder SUB XXX,1.
;--------
!MOV [v_Var1],12345
!DEC [v_Var1]
MessageRequester("DEC", "DEC 12345 = " + Str(Var1))
!MOV [v_Var2],12345
!INC [v_Var2]
MessageRequester("INC", "INC 12345 = " + Str(Var2))
;--------

; Fr die Multiplikation stehen uns die Befehle MUL (vorzeichenlos) und IMUL (vorzeichenbehaftet)
; zur Verfgung. Fangen wir mit dem einfacheren MUL an:
; Syntax: MUL Operand. Beeinflusste Flags: C, P, A, Z, S, O
; Operand kann ein Register oder eine Variable sein.
; Von den Flags werden nur Carry und Overflow definiert gesetzt und knnen ausgewertet werden,
; die anderen sind undefiniert und drfen nicht ausgewertet werden.
; Da fllt uns sofort auf: Multiplikation mit nur einem Operanden? Da fehlt doch einer! Nein,
; denn der Multiplikator ist immer der Akkumulator und darf nicht in der Befehlszeile angegeben
; werden. Eigentlich heisst es also z.B. "MUL EAX,Operand" fr eine vorzeichenlose 32-Bit-
; Multiplikation. Die Gre des Operanden bestimmt, welches Akkumulator-Teil-Register verwendet
; wird und wo das Ergebnis landet. Es knnen immer nur gleichgroe Werte miteinander 
; multipliziert werden, also nur Byte mit Byte, nur Word mit Word und nur DWord mit DWord.
; Es muss auch bercksichtigt werden, da das Ergebnis nicht mehr in die Operanden-Register-
; Gre passt: Byte mal Byte kann ja auch grer sein als 255 (was noch in ein Byte passen
; wrde). Daher gilt folgende Zuordnung:
;   Operanden-Gre     verwendeter Akkumulator-Teil     Ergebnis liegt in
;  ------------------------------------------------------------------------
;       Byte                       AL                           AX
;       Word                       AX                          DX:AX
;      DWord                      EAX                         EDX:EAX

; Die beiden Flags Carry und Overflow zeigen an, ob das Ergebnis noch in die Operanden-Gre
; gepasst hat, also AH (bei Byte), DX (bei Word) oder EDX (bei DWord) Null sind. Wenn ja, sind
; beide Flags Null, sonst Eins. Damit kann man ermitteln, ob der hherwertige Teil des
; Ergebnisses ausgewertet werden muss oder nicht (Stichwort Optimierung!).  
;--------
Global Var4.b      ;Byte-Variable
Global Var5.w      ;Word-Variable 
!MOV [v_Var4],4     ;Var4 bestimmt, ob AL, AX oder EAX bei MUL verwendet wird
!MOV al,5           ;oder AX,5 oder EAX,5; es wird nur AL verwendet!
!MUL [v_Var4]       ;also AL x Var4
!MOV [v_Var5],ax    ;hier MUSS AX stehen 
MessageRequester("Byte-MUL", "4 x 5 = " + StrU(Var5,#PB_Word))
;--------
Global Var6.w      ;Word-Variable
!MOV bx,444         ;BX bestimmt, ob AL, AX oder EAX bei MUL verwendet wird
!MOV eax,555        ;oder EAX,555; es wird nur AX verwendet!
!MUL bx             ;also AX x BX
!MOV [v_Var5],dx    ;hier MUSS DX stehen; High-Word des Ergebnisses 
!MOV [v_Var6],ax    ;hier MUSS AX stehen; Low-Word des Ergebnisses 
Var1 = (Var5 << 16) + (Var6 & $ffff) 
MessageRequester("Word-MUL", "444 x 555 = " + StrU(Var1,#PB_Long))
;--------
!MOV ebx,444444
!MOV eax,555555
!MUL ebx            ;also EAX x EBX
!MOV dword[v_Var3],eax       ;hier MUSS EAX stehen; Low-DWord des Ergebnisses
!MOV dword[v_Var3+4],edx     ;hier MUSS EDX stehen; High-DWord des Ergebnisses
MessageRequester("DWord-MUL", "444444 x 555555 = " + StrU(Var3,#PB_Quad))
;--------
; Ist bei den MUL-Operanden das hchstwertige Bit gesetzt wird dieses nicht als Vorzeichen,
; sondern als ganz normales Werte-Bit interpretiert:
;--------
!MOV ebx,-444444
!MOV eax,555555
!MUL ebx            ;also EAX x EBX
!MOV dword[v_Var3],eax       ;hier MUSS EAX stehen; Low-DWord des Ergebnisses
!MOV dword[v_Var3+4],edx     ;hier MUSS EDX stehen; High-DWord des Ergebnisses
MessageRequester("DWord-MUL mit 'negativer' Zahl ", "-444444 x 555555 = " + Str(Var3) + #LFCR$ + "denn -444444 wird auf Grund des Vorzeichen-Bits als 4294522852 interpretiert!")
;--------
; Weiter ist zu beachten, da bei der Byte- und Word-Multiplikation der obere Bereich (Bits 16
; bis 31) der verwendeten Register nicht verndert wird. Also anstelle von AX einfach EAX als
; Ergebnis-Register auszuwerten klappt nur, wenn diese Bits vorher schon auf Null gesetzt waren!  

; Die vorzeichenbehaftete Multiplikation IMUL macht das Operanden-Treiben verrckt: IMUL gibt
; es mit einem, mit zwei und mit drei Operanden!
; Fangen wir mal mit einem an:
; Syntax: IMUL Operand. Beeinflusste Flags: C, P, A, Z, S, O     
; Operand kann ein Register oder eine Variable sein.
; Von den Flags werden nur Carry und Overflow definiert gesetzt und knnen ausgewertet werden,
; die anderen sind undefiniert und drfen nicht ausgewertet werden.
; Und man ahnt es schon: Der "fehlende" zweite Operand ist natrlich wieder der Akkumulator.
; Die Zuordnungen Operanden und Ergebnis sind wie bei MUL (s.o.).
; Da oben bei MUL die Multiplikation mit -444444 in die Hose ging, probieren es mal mit IMUL:
;--------
!MOV ebx,-444444
!MOV eax,555555
!IMUL ebx           ;also EAX x EBX
!MOV dword[v_Var3],eax       ;hier MUSS EAX stehen; Low-DWord des Ergebnisses
!MOV dword[v_Var3+4],edx     ;hier MUSS EDX stehen; High-DWord des Ergebnisses
MessageRequester("DWord-IMUL mit negativer Zahl ", "-444444 x 555555 = " + Str(Var3))
;--------
; Na bitte, geht doch! Also: MUL = vorzeichenlos, IMUL = vorzeichenbehaftet. IMUL arbeitet
; natrlich auch mit "nur-positiven-"Zahlen korrekt und nicht langsamer als MUL, aber der 
; Wertebereich ist bei MUL grsser (doppelt so gro). Die Bedeutung des Carry- und Overflow-
; Flags ist wie bei MUL.

; Nun IMUL mit zwei Operanden:
; Syntax: IMUL Operand1, Operand2. Flags usw. wie oben, Carry und Overflow sind jetzt aber
; richtig wichtig!
; Operand1 mu ein 16-oder 32-Bit-Register sein, Operand2 kann eine gleichgrosse Variable,
; ein gleichgrosses Register oder eine Konstante (ein direkter Zahlenwert) sein. Die Konstante
; darf kein grsserer Typ als Operand1 sein; also wenn Operand1 vom Typ Word ist, darf die
; Konstante nur Byte-oder Word-Grsse haben. 
; Jetzt kommts: Das Ergebnis wird nur im 1.Operanden abgespeichert! Der wird also berschrieben
; (kann man noch verschmerzen), aber wenn das Ergebnis grsser ist als in das Register passt,
; steht man ziemlich doof da. Es mssen unbedingt die Flags Carry oder Overflow ausgewertet
; werden, um sicher zu gehen, da das Ergebnis richtig ist:   
;--------
AString$ = "Ergebnis ist in Ordnung!"
!MOV ebx,-444444     
!MOV ecx,5555
!IMUL ebx,ecx           ;also EBX x ECX
!MOV [v_Var1],ebx       ;hier MUSS EBX stehen; das Ergebnis steht ja in Operand1 
!JNC @f                 ;Carry wird gesetzt, wenn das Ergebnis nicht mehr in Operand1 passt!
AString$ = "Ergebnis ist grosser Mist!"     ;String ndern
!@@:                    ;anonymes Label, s.o.
MessageRequester("DWord-IMUL mit zwei Operanden", "Die Werte ergeben : " + Str(Var1) + #CRLF$ + AString$)
;--------
; Zum Test die Werte oben (z.B. EBX) ndern!
; Der Sprung JNC wird auch nach MOV... noch richtig ausgefhrt, da MOV selbst keine Flags
; verndert. JNO (Auswertung Overflow) wrde auch funktionieren.
; Also: Aufpassen! Wer auf Nummer Sicher gehen will (oder mu), nimmt die Variante mit nur
; einem Operanden; ab dem Pentium ist dessen Ausfhrungs-Geschwindigkeit kaum langsamer als
; die der anderen Varianten.  

; Fehlt noch IMUL mit drei Operanden:
; Syntax: IMUL Operand1, Operand2, Operand3. Flags usw. wie oben, Carry und Overflow sind auch
; hier richtig wichtig!
; Operand1 mu ein 16-oder 32-Bit-Register sein, Operand2 kann eine gleichgrosse Variable oder
; ein gleichgrosses Register sein und Operand3 mu eine Konstante (ein direkter Zahlenwert)
; passender Grsse sein (s.o.). Multipliziert werden die Operanden 2 und 3 miteinander, das Ergebnis
; wird in Operand1 gespeichert. Es gelten also die selben Bemerkungen wie fr IMUL mit zwei
; Operanden:
;--------
AString$ = "Ergebnis ist in Ordnung!"
!MOV ebx,5555555
!IMUL edx,ebx,-123      ;also EBX x -123
!MOV [v_Var1],edx       ;hier MUSS EDX stehen; Ziel ist ja Operand1 
!JNC @f                 ;Carry wird gesetzt, wenn das Ergebnis nicht mehr in Operand1 passt!
AString$ = "Ergebnis ist grosser Mist!"     ;String ndern
!@@:                    ;anonymes Label, s.o.
MessageRequester("DWord-IMUL mit zwei Operanden", "Die Werte ergeben : " + Str(Var1) + #CRLF$ + AString$)
;--------
; Auch hier gilt wieder: Einfach verschieden Werte ausprobieren!

; Kommen wir zur Division: Da gibt es DIV (vorzeichenlos) und IDIV (vorzeichenbehaftet). 
; Aber Achtung: Die Divisions-Programmierung ist sehr fehlertrchtig; man muss gut aufpassen. 
; Syntax: DIV Operand. Beeinflusste Flags: C, P, A, Z, S, O
; Syntax: IDIV Operand. Beeinflusste Flags: C, P, A, Z, S, O  
; Wichtig: Die genannten Flags werden zwar verndert (bzw. knnten es sein), ihr Status ist
; aber undefiniert und keines darf fr irgendwelche Zwecke ausgewertet werden!
; hnlich wie bei MUL wird bei beiden Divisions-Befehlen nur ein Operand angegeben, denn der
; Dividend ist immer der Akkumulator (ab Word-Grsse zusammen mit dem Data-Register) und darf
; nicht in der Befehlszeile angegeben werden. Eigentlich heisst es also z.B. "DIV EDX:EAX,Operand"
; fr eine vorzeichenlose 32-Bit-Division. 
; Operand kann ein Register oder eine Variable sein.
; Es gelten folgende Zuordnungen:
;   Operanden-Gre             Dividend                Quotient            Rest
;  ------------------------------------------------------------------------------
;       Byte                       AX                      AL                AH
;       Word                     DX:AX                     AX                DX
;      DWord                    EDX:EAX                   EAX               EDX

; Da ja nur Integer-Divisionen ausgefhrt werden, taucht hier auch ein (mglicher) Rest auf.
;--------

; Gem obiger Tabelle erfolgt die Word-(DWord-)Division auch mit dem (E)DX-Register als Teil
; des Dividenden. Der Programmierer ist jetzt selbst dafr verantwortlich, in (E)DX den
; richtigen Wert zu laden. Der Prozessor kann nicht wissen, das wir fr unser folgendes Werte-
; Beispiel das (E)DX- Register eigentlich gar nicht als Dividenden-Teil bruchten. Wird (E)DX
; nicht fr den Dividenden-Wert bentigt, mu es explizit auf Null gesetzt werden:
;--------
!MOV edx,0              ;mit dem spter erluterten Befehl XOR gehts (evtl.) etwas schneller
!MOV eax,12345
!MOV [v_Var1],678
!DIV [v_Var1]           ;EDX:EAX / Var1
!MOV [v_Var1],eax       ;EAX = Ergebnis
!MOV [v_Var2],edx       ;EDX = Rest
MessageRequester("DIV ", "12345 / 678 = " + StrU(Var1,#PB_Long) + "  Rest " + StrU(Var2,#PB_Long))    
;--------
; Falls hier jemand schon etwas von Befehlen wie CWD oder CDQ gehrt hat: Die haben bei der
; vorzeichenlosen Division DIV nichts zu suchen! Werden wir bei IDIV noch sehen.                      
; Da DIV vorzeichenlos arbeitet, wrde beispielsweise bei "MOV eax,-12345" fr EAX der Wert
; 4294954951 in die Rechnung eingehen (4294954951 = Unsigned -12345). 
; brigens, Rest ist nichts weiter als das von PB bekannte Modulo:
;--------
MessageRequester("PB-Modulo ", "Modulo 12345 / 678 = " + Str(12345 % 678)) 
;--------
 
; Das von einer Division durch Null die Finger zu lassen ist, drfte klar sein. Aber es lauern
; noch weitere Fallen:
;--------
Global Var7.b
!MOV ax,12345
!MOV bl,88
!DIV bl                 ;AX / BL
!MOV [v_Var4],al        ;AL = Ergebnis
!MOV [v_Var7],ah        ;AH = Rest
MessageRequester("DIV ", "Die Werte ergeben : " + StrU(Var4,#PB_Byte) + "  Rest " + StrU(Var7,#PB_Byte))    
;--------
; Das funktioniert erst mal, aber wenn BL von 88 auf z.B. 8 verringert wird, sehen wir: Nichts!
; Obwohl der Code mit BL=8 auch absolut korrekt ist (den Regeln entspricht), passt das Ergebnis
; nicht mehr in das Ergebnis-Register AL und es wird der Interrupt 0 ausgelst. Folge: Programm-
; Abbruch! Also: Vorher nachdenken und die Variante nehmen, bei der das "schlimmste" zu
; erwartende Ergebnis in das Ergebnis-Register passt.

; Zur vorzeichenbehafteten Division mit IDIV:
;--------
!MOV ax,12345           ;auch mal testen mit -12345
!CWD                    ;erst nach dem Laden von AX wirksam! 
!MOV [v_Var5],-678
!IDIV [v_Var5]          ;DX:AX / Var5
!MOV [v_Var5],ax        ;AX = Ergebnis
!MOV [v_Var6],dx        ;DX = Rest
MessageRequester("IDIV ", "12345 / -678 = " + Str(Var5) + "  Rest " + Str(Var6))    
;--------
; Ganz wichtig: Das Vorzeichen des Restes ist das gleiche wie das des Dividenden (hier AX)!
; Das Ergebnis-Vorzeichen wird entsprechend den blichen Rechen-Regeln bestimmt. 
; Neu ist der Befehl CWD (Convert Word to Double-Word). Er wird so wie oben eingesetzt (ohne
; Operanden) und sorgt hier fr das "richtige" Laden des DX-Registers. Bei den obigen Beispielen
; fr DIV (zur Erinnerung: vorzeichenlos) haben wir DX auf Null gesetzt, aber bei IDIV bestimmt
; das Register (E)DX das Vorzeichen des Dividenden (bis auf die Byte-Version). CWD besetzt alle
; Bits von DX mit dem Vorzeichen-Bit von AX. Bei DIV wre das fatal, wenn von AX das hchste Bit
; gesetzt wre, deshalb wurde dort direkt DX auf Null gesetzt (Inhalt von (E)DX wird als High-Wert
; des Dividenden benutzt). 
; Das "Double-Word" bei CWD bezieht sich also nicht auf ein 32-Bit-Register, sondern auf die
; beiden 16-Bit-Register AX und DX zusammen.
; Es gibt in diesem Zusammenhang 4 Convert-Befehle:
;        Befehl       kopiert hchstwertiges Bit von       in alle Bits von Register 
;       -----------------------------------------------------------------------------
;         CBW                      AL                                 AH
;         CDQ                     EAX                                EDX 
;         CWD                      AX                                 DX
;        CWDE                      AX                                EAX, Bits 16 bis 31 

; Da die Problematik mit dem (bisher nicht genutzten) High-Part des Dividenden bei DIV und
; IDIV extrem wichtig ist, habe ich ich sie hier mal sehr in den Vordergrund gestellt. Aber
; natrlich kann der High-Part auch "richtig" mit Werten genutzt werden, also der Dividend
; grssere Werte annehmen:
;--------
!MOV eax,123456789      ;Low-Part des Dividenden
!MOV edx,9876           ;High-Part des Dividenden; also kein Convert o.. ntig  

!MOV dword[v_Var3],eax  ;dieser Teil hier nur fr die Anzeige unten!
!MOV dword[v_Var3+4],edx 

!MOV dword[v_Var1],-67890
!IDIV [v_Var1]          ;EDX:EAX / Var1
!MOV [v_Var1],eax       ;EAX = Ergebnis
!MOV [v_Var2],edx       ;EDX = Rest
MessageRequester("IDIV mit grossen Werten", Str(Var3) + " / -67890 = " + Str(Var1) + "  Rest " + Str(Var2))     
;--------
; Zum Test kann das Vorzeichen von EDX gendert werden.
; Wenn im Beispiel EDX wesentlich vergrssert wird schlgt wieder der Interrupt 0 zu, da das
; Ergebnis dann nicht mehr in ein 32-Bit-Register passt. Da helfen nur spezielle Routinen
; oder der Umstieg auf Fliess-Komma. Interrupt-(bzw. Exception-)Auswertung ist ein Thema fr
; sich und wrde hier viel zu weit fhren. Also: Darauf achten, da das Ergebnis nicht zu
; gro werden kann!

; Fr die Multiplikation und Division mit 2^n-Werten (also 2, 4, 8, 16 usw) kann man die
; schnellen Bit-Shift-Befehle SHL (Shift Left; schiebe nach links) und SHR (Shift Right;
; schiebe nach rechts) verwenden:
; Syntax: SHL Operand1, Operand2. Beeinflusste Flags: C, P, A, Z, S, O (wobei A undefiniert ist).
; Syntax: SHR Operand1, Operand2. Flags wie SHL.
; Operand1 kann ein Register oder Variable sein, Operand2 gibt die Anzahl der Verschiebungen
; an und kann eine 8-Bit-Konstante oder das CL-Register sein.
; Verschoben wird in Operand1 bit-weise, die "frei" werdenden Bits am anderen "Ende" werden mit
; Null-Bits aufgefllt:
;--------
!MOV eax,1234567
!SHR eax,2              ;2 Verschiebungen nach rechts entsprechen einer Division durch 4 (2^2=4) 
!MOV [v_Var1],eax
MessageRequester("SHR", "1234567 Shift Rechts 2 ergibt : " + Str(Var1))
;--------
; Und wie man sieht, ist hier erstmal nichts mit einer Rest-Angabe. Das Ergebnis ist immer
; exakt (Rest=0) oder abgerundet. Wenn das nicht strt, ist Shift Rechts eine feine Sache.
; Da bei den Shift-Befehlen das zuletzt aus Operand1 herausgeschobene Bit im Carry-Flag
; landet, knnte man Code erstellen, der das fr eine Rest-Ermittlung auswertet. Aber: Dann
; kann man auch gleich wieder die "richtige" Division verwenden.
;--------
!MOV eax,1234567
!MOV cl,2               ;hier benutzen wir mal das Register CL 
!SHL eax,cl             ;2 Verschiebungen nach links entsprechen einer Multiplikation mit 4 (2^2=4) 
!MOV [v_Var1],eax
MessageRequester("SHL", "1234567 Shift Links 2 ergibt : " + Str(Var1))
;-------- 

; Bei Shift Links besteht die Gefahr, das Bits nach links "herausgeschoben" werden und somit
; das Ergebnis Muster ohne Wert wird. Da diese Bits aber ber das Carry-Flag laufen kann man
; das so testen:
;--------
!MOV eax,1234567
!@@:
!MOV edx,eax            ;EAX sichern in EDX
!SHL eax,1
!JC @f                  ;beim ersten herausgeschobenen Bit erfogt Abbruch
!JMP @b                 ;zurck zum nchsten Shift
!@@:                    ;Abbruch
!MOV [v_Var1],edx
MessageRequester("SHL", "Der grsste (richtige) erreichbare Wert ist : " + StrU(Var1,#PB_Long))
;--------
; Zur Sicherheit msste eigentlich der Startwert von EAX auf Null getestet werden, denn in
; diesem Fall htte man eine prima Endlos-Schleife!
; Zu beachten ist, das im Carry-Flag immer das zuletzt herausgeschobene Bit landet. Lauten
; die 3 hchsten Bits z.B. 010 und wir shiften 3-mal in einem Rutsch nach links, ist nach
; dieser Operation das Carry-Flag nicht gesetzt (letzte Null ist dort drin)! Die mittlere 1
; landet im Nirwana und entzieht sich unserer Auswertung.
                      
; Die beiden o.g. Befehle fllen ja die frei werden Stellen mit Null-Bits auf; aber es geht
; auch anders:
; Syntax: SHLD Operand1, Operand2, Operand3. Beeinflusste Flags: C, P, A, Z, S, O (wobei A
; und O (s.u.) undefiniert sind).
; Syntax: SHRD Operand1, Operand2, Operand3. Flags wie SHLD.
; Operand1 kann ein 16- oder 32-bit-Register oder Variable sein, Operand2 muss ein Register
; gleicher Grsse sein. Operand3 gibt die Anzahl der Verschiebungen an und kann eine 8-Bit-
; Konstante oder das CL-Register sein.
; Neu ist hier der mittlere Operand2, dessen Bits beim Shiften in Operand1 "nachgeschoben"
; werden; Operand2 bleibt aber unverndert. Bei SHLD fngt das Nachschieben mit dem MSB, bei
; SHRD mit dem LSB von Operand2 an.
; Nur wenn die Anzahl der Verschiebungen Eins ist, kann das Overflow-Flag ausgewertet werden;
; bei einem Vorzeichen-Wechsel wird es gesetzt.
;--------
!MOV [v_Var6],1100110011001100b   ;Word-Variable
!MOV ax,      1010101010101010b   ;frei gewhlt
!SHLD [v_Var6],ax,3               ;3-mal nach links verschieben = 3 obere Bits von AX
                                  ; nachschieben
MessageRequester("SHLD", "Das Bitmuster sieht so aus : " + RSet(Bin(Var6,#PB_Word),16,"0"))

!MOV [v_Var6],1100110011001100b
!MOV ax,      1010101010101010b
!SHRD [v_Var6],ax,3               ;3-mal nach rechts verschieben = 3 untere Bits von AX
                                 ; nachschieben
MessageRequester("SHRD", "Das Bitmuster sieht so aus : " + RSet(Bin(Var6,#PB_Word),16,"0"))
;--------

; Weitere Bit-Schiebe-Befehle sind SAL und SAR. SAL ist identisch mit SHL; SAR funktioniert
; wie SHR, aber das "verlustig gegangene" oberste Bit wird nicht durch ein Null-Bit ersetzt,
; sondern durch das vorher vorhandene MSB, also dem Vorzeichen bei vorzeichenbehafteter
; Interpretation:
;--------
!MOV eax,-1234567
!MOV cl,2 
!SAR eax,cl             ;2 Verschiebungen nach rechts entsprechen einer vorzeichenbehafteten 
;                         Division durch 4 (2^2=4) 
!MOV [v_Var1],eax
MessageRequester("SAR", "-1234567 SAR 2 ergibt : " + Str(Var1))
;-------- 
; Ist der Operand negativ, rundet SAR praktisch das Ergebnis gegen minus unendlich. EAX mal
; positiv machen!

; Weitere Bit-Schiebe-Befehle sind die Rotations-Befehle, bei denen das herausgeschobene Bit
; am anderen Ende wieder reingeschoben wird (also rotiert).
; Syntax: RCL Operand1, Operand2. Beeinflusstes Flag: Nur C
; Syntax: RCR Operand1, Operand2. Beeinflusstes Flag: Nur C
; Syntax: ROL Operand1, Operand2. Beeinflusste Flags: Keine
; Syntax: ROR Operand1, Operand2. Beeinflusste Flags: Keine
; Operand1 kann ein Register oder Variable sein, Operand2 gibt die Anzahl der Verschiebungen
; an und kann eine 8-Bit-Konstante oder das CL-Register sein.
; RCL rotiert nach links ber das Carry-Flag, d.h. das zuletzt herausrotierte Bit (das MSB)
; landet im Carry-Flag. Hereingeschoben wird der Inhalt des Carry-Flags, und hier lauert
; die Tcke: Beim ersten Schiebevorgang ist ja der Zustand dieses Flags noch zufllig, d.h.
; der Programmierer muss fr einen definierten Zustand sorgen, es sei denn, er will bewusst
; das Ergebnis einer vorhergehenden Operation (Carry-Flag betreffend) einfliessen lassen. 
; RCR rotiert nach rechts ber das Carry-Flag, d.h. das zuletzt herausrotierte Bit (das LSB)
; landet im Carry-Flag. Sonst wie RCL.

; ROL und ROR arbeiten analog, nur ohne irgendein Flag zu beeinflussen und ohne die
; Zwischenstation Carry-Flag:
;--------
Global Var8.l
!MOV [v_Var1],11110000111100001111000011110000b
!MOV eax,[v_Var1]
!ROL eax,3              ;3 Verschiebungen nach links; was dort rausfliegt, wird rechts direkt
                        ; wieder als LSB eingefgt     
!MOV [v_Var2],eax
!MOV eax,[v_Var1]
!ROR eax,3              ;3 Verschiebungen nach rechts; was dort rausfliegt, wird links direkt
                        ; wieder als MSB eingefgt
!MOV [v_Var8],eax
MessageRequester("Rotations-Befehle ROL und ROR", "Ausgangswert : " + RSet(Bin(Var1,#PB_Long),32,"0") + #LFCR$ + "ROL 3 ergibt :    " + RSet(Bin(Var2,#PB_Long),32,"0") + #LFCR$ + "ROR 3 ergibt :   " + RSet(Bin(Var8,#PB_Long),32,"0"))
;-------- 

;--------
!MOV [v_Var1],11110000111100001111000011110000b  ;hier nochmal definiert
!MOV eax,[v_Var1]
!CLC                    ;Neu: CLC lscht das Carry-Flag
!RCL eax,3              ;3 Verschiebungen nach links; was dort rausfliegt, landet im Carry-Flag;
                        ; vorher wird der alte Inhalt von Carry rechts als LSB eingefgt
!MOV [v_Var2],eax
!MOV eax,[v_Var1]
!STC                    ;Neu: STC setzt das Carry-Flag
!RCR eax,3              ;3 Verschiebungen nach rechts; was dort rausfliegt, landet im Carry-Flag;
                        ; vorher wird der alte Inhalt von Carry links als MSB eingefgt
!MOV [v_Var8],eax
MessageRequester("Rotations-Befehle RCL und RCR", "Ausgangswert : " + RSet(Bin(Var1,#PB_Long),32,"0") + #LFCR$ + "RCL 3 ergibt :    " + RSet(Bin(Var2,#PB_Long),32,"0") + #LFCR$ + "RCR 3 ergibt :   " + RSet(Bin(Var8,#PB_Long),32,"0"))
;-------- 
; Hier ist zu sehen, da das erste geschobene Bit dem zwangsweise gesetztem bzw. gelschtem
; Carry-Flag entspricht. 
; Zu beachten ist noch bei allen Schiebe-bzw.Rotations-Befehlen, da die Anzahl der Verschiebungen
; maximal 31 betrgt, da vom Operanden, der die Anzahl angibt, nur die ersten 5 Bits als Zhler
; Verwendung finden. So etwas wie "SHR Operand,123" wird ohne Fehlermeldung akzeptiert, es finden
; aber nur 27 Verschiebungen statt (123d = 01111011b, Bit 0 bis 4 = 27). Der maximal zulssige
; Wert fr den Zhler ist 255. 

; Mit CLC und STC kann man das Carry-Flag zwangsweise lschen bzw. setzen und sollte klar sein
; (CLC = Clear Carry und STC = Set Carry). Nebenbei: Fr das Carry-Flag gibt es sogar einen
; Toggle- Befehl: CMC (Complement Carry) invertiert das Flag. 

; Kommen wir zu weiteren Befehlen, bei denen die einzelnen Bits die Hauptrolle spielen:
; Syntax: NEG Operand1. Beeinflusste Flags: C, P, A, Z, S, O
; NEG negiert den Operanden, der ein Register oder Variable sein kann. Es wird das 
; Zweierkomplement gebildet; d.h. es wird die Operation "Null minus Operand1" ausgefhrt.
;--------
!MOV [v_Var1],123456
!NEG [v_Var1]
MessageRequester("NEG", "NEG 123456 ergibt : " + Str(Var1))
;--------
; Vordergrndig ist das Ergebnis ein schlichter Vorzeichenwechsel. Auf Bit-Ebene luft folgendes
; ab: Alle Bits von Operand1 werden invertiert (aus 0 wird 1 und umgekehrt) und anschliessend
; wird 1 dazu addiert.
; Syntax: NOT Operand1. Beeinflusste Flags: Keine
; NOT bildet das Einerkomplement von Operand1, der ein Register oder Variable sein kann.
; Alle Bits von Operand1 werden invertiert (aus 0 wird 1 und umgekehrt). 
;--------
!MOV [v_Var1],11110000111100001111000011110000b
!MOV [v_Var2],11110000111100001111000011110000b  ;fr Anzeige dasselbe nochmal
!MOV [v_Var8],11110000111100001111000011110000b
!NOT [v_Var1]
!NEG [v_Var8]
MessageRequester("NOT und NEG", "Ausgangswert : " + RSet(Bin(Var2,#PB_Long),32,"0") + #LFCR$ + "NOT ergibt :      " + RSet(Bin(Var1,#PB_Long),32,"0") + #LFCR$ + "NEG ergibt :      " + RSet(Bin(Var8,#PB_Long),32,"0"))  
;--------
; Bei NEG nicht durch die Nullen am Ende verwirren lassen; als Basis mal das Ergebnis von
; NOT nehmen und 1 addieren! 

; Syntax: AND Operand1, Operand2. Beeinflusste Flags: C, P, A, Z, S, O
; Die Operanden knnen Register oder Variable (nicht beide) sein, Operand2 kann auch eine
; Konstante sein.
; Die Flags Carry und Overflow werden auf Null gesetzt, Auxiliary ist undefiniert.
; AND fhrt ein bit-weises UND der beiden Operanden durch und schreibt das Ergebnis in Operand1:
;--------
!MOV [v_Var1],11110000111100001111000011110000b
!MOV eax,     11000000110000001100000011000000b
!MOV [v_Var2],eax
!AND eax,[v_Var1]
!MOV [v_Var8],eax
MessageRequester("AND", "Operand1 :    " + RSet(Bin(Var2,#PB_Long),32,"0") + #LFCR$ + "Operand2 :    " + RSet(Bin(Var1,#PB_Long),32,"0") + #LFCR$ + "AND ergibt :   " + RSet(Bin(Var8,#PB_Long),32,"0"))  
;-------- 
; Es bleibt in Operand1 also nur das Bit gesetzt, das in beiden Operanden gesetzt ist. Operand2
; bleibt unverndert.

; Syntax: TEST Operand1, Operand2. Beeinflusste Flags: C, P, A, Z, S, O
; Die Operanden knnen Register oder Variable (nicht beide) sein, Operand2 kann auch eine
; Konstante sein.
; Die Flags Carry und Overflow werden auf Null gesetzt, Auxiliary ist undefiniert.
; TEST fhrt genau wie AND ein bit-weises UND der beiden Operanden durch, verndert aber
; keinen Operanden! Der Programmierer wertet lediglich die Flag-Zustnde nach TEST aus und
; leitet entsprechende Operationen ein (meist bedingte Sprnge). 
;--------
!MOV [v_Var1],10011b
!TEST [v_Var1],11b      ;testet, ob von Var1 eines der Bits 0 oder 1 gesetzt ist
!JZ @f                  ;springe, wenn beide Bits nicht gesetzt sind (Ergebnis = Null)
MessageRequester("TEST", "Mindestens eines der Bits 0 und 1 ist gesetzt!")
!@@:
;--------
; Mit TEST kann man z.B. ermitteln, ob ein Wert negativ ist oder nicht:
;--------
!MOV [v_Var1], 11110000111100001111000011110000b
!TEST [v_Var1],10000000000000000000000000000000b ; nur Test auf das MSB (Vorzeichen-Bit)
!JNS @f                 ;springe, wenn das MSB nicht gesetzt ist
MessageRequester("TEST", "Das Vorzeichen-Bit ist gesetzt!")
!@@:
;--------
; Anstelle von JNS knnte hier auch wieder JZ verwendet werden, aber mit JNS signalisiert
; der Programmierer, das er nicht auf Null, sondern auf das Vorzeichen testen wollte. Das
; macht Code lesbarer! 

; Syntax: OR Operand1, Operand2. Beeinflusste Flags: C, P, A, Z, S, O
; Die Operanden knnen Register oder Variable (nicht beide) sein, Operand2 kann auch eine
; Konstante sein.
; Die Flags Carry und Overflow werden auf Null gesetzt, Auxiliary ist undefiniert.
; OR fhrt ein bit-weises ODER der beiden Operanden durch und schreibt das Ergebnis in Operand1:
;--------
!MOV [v_Var1],11110000111100001111000011110000b
!MOV eax,     11000100110000001100010011000100b
!MOV [v_Var2],eax
!OR eax,[v_Var1]
!MOV [v_Var8],eax
MessageRequester("OR", "Operand1 :    " + RSet(Bin(Var2,#PB_Long),32,"0") + #LFCR$ + "Operand2 :    " + RSet(Bin(Var1,#PB_Long),32,"0") + #LFCR$ + "OR ergibt :     " + RSet(Bin(Var8,#PB_Long),32,"0"))  
;-------- 
; In Operand1 wird das Bit gesetzt, das in einen oder beiden der Operanden gesetzt ist. Operand2
; bleibt unverndert.

; Syntax: XOR Operand1, Operand2. Beeinflusste Flags: C, P, A, Z, S, O
; Die Operanden knnen Register oder Variable (nicht beide) sein, Operand2 kann auch eine
; Konstante sein.
; Die Flags Carry und Overflow werden auf Null gesetzt, Auxiliary ist undefiniert.
; XOR fhrt ein bit-weises EXCLUSIV-ODER der beiden Operanden durch und schreibt das Ergebnis
; in Operand1:
;--------
!MOV [v_Var1],11110000111100001111000011110000b
!MOV eax,     11000100110000001100010011000100b
!MOV [v_Var2],eax
!XOR eax,[v_Var1]
!MOV [v_Var8],eax
MessageRequester("XOR", "Operand1 :    " + RSet(Bin(Var2,#PB_Long),32,"0") + #LFCR$ + "Operand2 :    " + RSet(Bin(Var1,#PB_Long),32,"0") + #LFCR$ + "XOR ergibt :  " + RSet(Bin(Var8,#PB_Long),32,"0"))  
;-------- 
; In Operand1 wird nur das Bit gesetzt, das nur in einen der beiden Operanden gesetzt ist.
; Operand2 bleibt unverndert. Deshalb wird XOR gern benutzt, um ein Register auf Null zu
; setzen. Anstelle von z.B. MOV EAX,0 wird XOR EAX,EAX verwendet, was schneller ist:
;--------
!MOV eax,11000100110000001100010011000100b  ;irgendein beliebiger Wert
!MOV [v_Var1],eax                           ;nur fr Anzeige!
!XOR eax,eax                                ;setzt EAX auf Null
!MOV [v_Var8],eax
MessageRequester("XOR, um ein Register auf Null zu setzen", "Operand1 :    " + RSet(Bin(Var1,#PB_Long),32,"0") + #LFCR$ + "Operand2 :    " + RSet(Bin(Var1,#PB_Long),32,"0") + #LFCR$ + "XOR ergibt :  " + RSet(Bin(Var8,#PB_Long),32,"0"))   
;-------- 

; Weitere Bit-Befehle sind die Bit-Scan- und Bit-Test-Befehle:
; Syntax: BSF Operand1, Operand2. Beeinflusste Flags: C, P, A, Z, S, O. Nur Z ist definiert,
; der Rest ist undefiniert.
; Operand1 muss ein 16-oder 32-Bit-Register sein, Operand2 kann ein Register oder Variable
; gleicher Grsse wie Operand1 sein.
; BSF (Bit Scan Forward) durchsucht Operand2, beginnend bei Bit0, nach dem ersten gesetztem Bit
; und schreibt dessen Position in Operand1 sowie lscht das Zero-Flag:
;--------
!MOV edx,11000100110000001100010011000100b  ;Operand2
!MOV eax,1234567890     ;beliebiger, zuflliger Wert, der in EAX (Operand1) stehen kann
!BSF eax,edx
!JZ @f                  ;springe, wenn Operand2 Null 
!MOV [v_Var1],eax
MessageRequester("BSF", "Das erste gesetzte Bit ist an Stelle " + Str(Var1))
!@@:
;--------
; Beachten, das Bit-Zhlungen immer mit Wert Null beginnen! 
; Hat Operand2 den Wert Null (oben testen!), wird Operand1 nicht verndert und das Zero-Flag
; wird gesetzt.

; Syntax: BSR Operand1, Operand2. Beeinflusste Flags: C, P, A, Z, S, O. Nur Z ist definiert,
; der Rest ist undefiniert.
; BSR (Bit Scan Reverse) durchsucht Operand2, beginnend beim hchstwertigen Bit (MSB), nach
; dem ersten gesetztem Bit und schreibt dessen Position in Operand1 sowie lscht das Zero-Flag.
; Sonst wie BSF.
;--------
!MOV edx,11000100110000001100010011000100b  ;Operand2
!MOV eax,1234567890     ;beliebiger, zuflliger Wert, der in EAX (Operand1) stehen kann
!BSR eax,edx
!JZ @f                  ;springe, wenn Operand2 Null
!MOV [v_Var1],eax
MessageRequester("BSR", "Das erste gesetzte Bit ist an Stelle " + Str(Var1))
!@@:
;--------
; Im praktischen Einsatz sollte immer das Zero-Flag ausgewertet werden!
  
; Syntax: BT Operand1, Operand2. Beeinflusste Flags: C, P, A, Z, S, O. Nur C ist definiert,
; der Rest ist undefiniert.
; Operand1 kann ein 16-oder 32-Bit-Register oder Variable sein, Operand2 kann ein Register
; gleicher Grsse wie Operand1 oder eine 8-Bit-Konstante sein.
; Operand2 gibt die Position des Bits von Operand1, welches in das Carry-Flag kopiert werden
; soll, an. Die Zhlung beginnt bei Bit0 (LSB):
;--------
BT_String$ = "nicht "
!MOV eax,11000100110000001100010011000100b  ;Operand1
;                                ^<-- das hier ist Bit7!  
!BT eax,7               ;kopiere Bit7 von Operand1 in das Carry-Flag
!JNC @f                 ;springe, wenn das Carry-Flag nicht gesetzt ist, also Bit7 Null ist  
BT_String$ = ""
!@@:
MessageRequester("BT", "Das untersuchte Bit war " + BT_String$ + "gesetzt!")
;--------
; BT (Bit Test) verndert keinen Operanden.

; Syntax: BTC Operand1, Operand2. Beeinflusste Flags: Wie BT
; Operanden wie BT. ABER: BTC verndert Operand1! BTC steht fr "Bit Test and Complement",
; das mit Operand2 angegebene Bit von Operand1 wird nicht nur in das Carry-Flag kopiert,
; sondern es wird anschliessend auch in Operand1 invertiert ("komplementiert")! 
;--------
BT_String$ = "nicht "
!MOV eax,11000100110000001100010011000100b  ;Operand1
!MOV edx,7              ;Operand2 mal als Register
!BTC eax,edx            ;kopiere Bit7 von Operand1 in das Carry-Flag und invertiere es in Operand1
!PUSH eax               ;Register EAX retten, wird nochmal gebraucht! PB knnte es mit den
                        ; nachfolgenden Befehlen verndern. PUSH verndert keine Flags!
!JNC @f                 ;springe, wenn das Carry-Flag nicht gesetzt ist, also Bit7 Null war  
BT_String$ = ""
!@@:
MessageRequester("BTC", "Das untersuchte Bit war " + BT_String$ + "gesetzt!")
BT_String$ = "nicht "   ;gleiche Voraussetzungen wie oben
!POP eax                ;EAX zurckholen 
!BT eax,7               ;das gleiche Bit nochmal abfragen; hier tut es jetzt BT
!JNC @f 
BT_String$ = ""
!@@:
MessageRequester("BTC", "Das untersuchte Bit ist jetzt " + BT_String$ + "gesetzt!")
;--------
; Man sieht, da das entsprechende Bit in Operand1 getoggelt ("gekippt") wurde.
  
; Syntax: BTR Operand1, Operand2. Beeinflusste Flags: Wie BT
; Operanden wie BT. ABER: BTR verndert Operand1! BTR steht fr "Bit Test and Reset",
; das mit Operand2 angegebene Bit von Operand1 wird nicht nur in das Carry-Flag kopiert,
; sondern es wird anschliessend auch in Operand1 gelscht! 
;--------
BT_String$ = "nicht "
!MOV eax,11000100110000001100010011000100b  ;Operand1
!BTR eax,7              ;kopiere Bit7 von Operand1 in das Carry-Flag und lsche es in Operand1
!PUSH eax               ;Register EAX retten, wird nochmal gebraucht! PB knnte es mit den
                        ; nachfolgenden Befehlen verndern. PUSH verndert keine Flags!
!JNC @f                 ;springe, wenn das Carry-Flag nicht gesetzt ist, also Bit7 Null war  
BT_String$ = ""
!@@:
MessageRequester("BTR", "Das untersuchte Bit war " + BT_String$ + "gesetzt!")
BT_String$ = "nicht "   ;gleiche Voraussetzungen wie oben
!POP eax                ;EAX zurckholen 
!BT eax,7               ;das gleiche Bit nochmal abfragen; hier tut es jetzt BT
!JNC @f 
BT_String$ = ""
!@@:
MessageRequester("BTR", "Das untersuchte Bit ist jetzt " + BT_String$ + "gesetzt!")
;--------
; Zum Test kann oben in EAX das Bit7 gelscht werden, dann sieht man den Unterschied zu BTC.
  
; Syntax: BTS Operand1, Operand2. Beeinflusste Flags: Wie BT
; Operanden wie BT. ABER: BTS verndert Operand1! BTS steht fr "Bit Test and Set",
; das mit Operand2 angegebene Bit von Operand1 wird nicht nur in das Carry-Flag kopiert,
; sondern es wird anschliessend auch in Operand1 gesetzt! 
;--------
BT_String$ = "nicht "
!MOV [v_Var1],11000100110000001100010011000100b  ;Operand1 mal als Variable
!BTS [v_Var1],7         ;kopiere Bit7 von Operand1 in das Carry-Flag und setze es in Operand1
!JNC @f                 ;springe, wenn das Carry-Flag nicht gesetzt ist, also Bit7 Null war  
BT_String$ = ""
!@@:
MessageRequester("BTS", "Das untersuchte Bit war " + BT_String$ + "gesetzt!")
BT_String$ = "nicht "   ;gleiche Voraussetzungen wie oben
!BT [v_Var1],7          ;das gleiche Bit nochmal abfragen; hier tut es jetzt BT
!JNC @f 
BT_String$ = ""
!@@:
MessageRequester("BTS", "Das untersuchte Bit ist jetzt " + BT_String$ + "gesetzt!")
;--------
; Zum Test kann oben in Var1 das Bit7 gelscht werden, dann sieht man den Unterschied zu BTR. 

; Zum Setzen eines ganzen Bytes oder einer 8-Bit-Variable mit den Werten 0 oder 1 in
; Abhngigkeit von den Flag-Zustnden gibt es die SETxx-Befehle:
; Syntax: SETxx Operand1. Beeinflusste Flags: Keine
; "xx" steht hier fr die gewnschte Auswertung der Flags genau wie bei den bedingten Sprngen
; (s.o.). Ist die Bedingung erfllt, erhlt Operand1 (8-Bit-Register oder -Variable) den 
; Wert Eins, ansonsten Null. Hier die bersicht: 

; Bedingter Setz-Befehl  |  ausgewertete Flags, setze Operand1 auf Eins wenn...                     
; ------------------------------------------------------------------------------
; SETA, SETNBE           |  Carry und Zero beide nicht gesetzt
; SETAE, SETNB           |  Carry nicht gesetzt
; SETB, SETNAE, SETC     |  Carry gesetzt
; SETBE, SETNA           |  Carry oder Zero gesetzt
; SETE, SETZ             |  Zero gesetzt
; SETG, SETNLE           |  Signum und Zero beide nicht gesetzt
; SETGE, SETNL           |  Signum nicht gesetzt
; SETL, SETNGE           |  Signum und Overflow ungleich
; SETLE, SETNG           |  Signum und Overflow ungleich oder Zero gesetzt
; SETNC                  |  Carry nicht gesetzt
; SETNE, SETNZ           |  Zero nicht gesetzt
; SETNO                  |  Overflow nicht gesetzt
; SETNP, SETPO           |  Parity nicht gesetzt
; SETNS                  |  Signum nicht gesetzt
; SETO                   |  Overflow gesetzt
; SETP, SETPE            |  Parity gesetzt
; SETS                   |  Signum gesetzt   

; Hier ein einfaches Beispiel:
;--------
!MOV [v_Var1],123456    ;beliebige Variable mit beliebigen Wert fllen
!SUB [v_Var1],200000    ;von Var1 200000 subtrahieren, wenn Var1 < 200000 ist, wird das
                        ; Carry-Flag gesetzt; auch mal Test mit 100000
!SETB [v_Var7]          ;wertet das Carry-Flag aus; wenn gesetzt, schreibe 1 in Var7, sonst 0
MessageRequester("SETB", Str(Var7))
;--------
; Wenn oben die bedingten Sprung-Bedingungen klar waren, ist dies hier auch klar.
; Gleiches gilt fr die bedingten Kopier-Befehle CMOVxx:
; Syntax: CMOVxx Operand1, Operand2. Beeinflusste Flags: Keine
; Operand1 muss ein 16-oder 32-Bit-Register sein, Operand2 kann ein Register oder Variable
; gleicher Grsse wie Operand1 sein.
; Ist die Bedingung erfllt, wird der MOV-Befehl ausgefhrt, sonst nicht: 

;--------
!MOV eax,1234567890     ;EAX mit beliebigen Wert laden
!MOV [v_Var1],123456    ;beliebige Variable mit beliebigen Wert fllen
!SUB [v_Var1],200000    ;von Var1 200000 subtrahieren, wenn Var1 < 200000 ist, wird das
                        ; Carry-Flag gesetzt; auch mal Test mit 100000
!CMOVB eax,[v_Var1]     ;wertet das Carry-Flag aus; wenn gesetzt, berschreibe EAX mit Var1
                        ; (dem Ergebnis der Subtraktion); sonst bleibt es unangetastet
!MOV [v_Var2],eax
MessageRequester("CMOVB", Str(Var2))
;--------
; Hier ist anzumerken, das in der (lteren) Literatur immer darauf hingewiesen wird, vor dem
; Einsatz von CMOVxx erstmal dessen Verfgbarkeit mittels des Befehls CPUID zu ermitteln. Aber
; ich behaupte mal, bei allen Intel- und AMD-Prozessoren der letzten 5 Jahre steht dieser Befehl
; zur Verfgung. 
; Aber wenn wir schon bei CPUID (CPU Identification) sind: 
; Syntax: CPUID. Keine Operanden. Beeinflusste Flags: Keine
; CPUID frdert alles das zu Tage, was der Hersteller an Informationen ber den Prozessor mit
; in Silizium gegossen hat. Das Ganze ist inzwischen sehr komplex geworden und mit fast jeder
; neuen Prozessor-Generation kommt Neues hinzu, Altes wird erweitert usw. Deshalb hier nur der
; Test, ob CMOVxx zu Verfgung steht:
;--------
; Den Test, ob CPUID berhaupt selbst zur Verfgung steht, schenke ich mir hier. Ab der Pentium-
; Generation gibt es ihn.
CMOV_String$ = "nicht "
!MOV eax,1              ;(ein) Aufruf-Parameter fr CPUID
!CPUID                  
!TEST edx,8000h         ;Test auf Bit15 von Register EDX, in dem CPUID hier Rckgabewerte liefert   
!JZ @f                  ;wenn gesetzt, steht der CMOVxx-Befehl zur Verfgung
CMOV_String$ = ""
!@@:
MessageRequester("CPUID - Test auf CMOVxx", "Der CMOVxx-Befehl steht " + CMOV_String$ + "zur Verfgung!")
;--------
; Nebenbei: CPUID ist auch ein sogenannter serialisierender Befehl, d.h. vor dessen Ausfhrung
; werden erst noch alle anstehenden Operationen innerhalb der CPU durchgefhrt.
; CPUID ist einer der (wenigen) internen CPU-Befehle, die auf allen Privileg-Stufen (diese
; legen die Zugriffs-Rechte fest) ausgefhrt werden knnen.
; Ist das Betriebs-System gndig (Windows ist es hier mal), kann auch der sonst privilegierte
; Befehl RDTSC (Read Time-Stamp Counter), der prima fr Zeitmessungen zu verwenden ist, verwendet
; werden:
; Syntax: RDTSC. Keine Operanden. Beeinflusste Flags: Keine 
; RDTSC liest den Zeit-Stempel-Zhler des Prozessors aus und speichert den Wert im Register-
; Paar EDX:EAX. Dieser Zhler wird mit jedem CPU-Takt inkrementiert und kann deshalb auch zur
; Ermittlung der momemtanen CPU-Takt-Frequenz benutzt werden:
;--------
!XOR eax,eax
!CPUID                  ;hier benutzen wir CPUID zum serialisieren; es soll nichts mehr stren  
!RDTSC                  ;Time-Stamp zum 1.mal auslesen 
!MOV [v_Var1], eax      ;Low-Wert zwischenspeichern 
!MOV [v_Var2], edx      ;High-Wert zwischenspeichern  
Delay(1000)             ;1 Sekunde warten (1000ms) 
!RDTSC                  ;Time-Stamp zum 2.mal auslesen 
!SUB eax,[v_Var1]       ;Differenz bilden 
!SBB edx,[v_Var2]       ;siehe SBB-Beispiel oben!
!MOV ecx, 1000          ;durch 1000 dividieren (1000ms) 
!DIV ecx
!MOV [v_Var1],eax       ;Taktfrequenz in Hertz 
Takt.d = Var1 / 1000    ;umwandeln in Float und GHz
  
If Takt > 1000
   Takt / 1000
   CPU$ = "CPU-Taktfrequenz = " + StrD(Takt,1) + " GHz"
  Else 
   CPU$ = "CPU-Taktfrequenz = " + StrD(Takt,0) + " MHz"
EndIf 

MessageRequester("CPU-Taktfrequenz", CPU$)
;--------
; Das soll es gewesen sein mit den Prozessor- und Betriebssystem-Nahen Befehlen. Derer gibt
; es eine ganze Menge, aber sie sind fr die "normale" Programmierung nicht weiter von Bedeutung;
; schon deshalb nicht, weil man normalerweise keinen Zugriff darauf hat. 

; Gebrauchen kann man aber folgende Befehle noch:
; Syntax: BSWAP Operand1. Beeinflusste Flags: Keine
; Operand1 mu ein 32-Bit-Register sein. BSWAP vertauscht die 4 Bytes von Operand1 so, da
; aus dem Little-Endian-Format das Big-Endian-Format wird und umgekehrt. Little-Endian-Format:
; Beim Abspeichern von 16-oder 32-Bit-Registern oder -Variablen wird das unterste (niedrigst-
; wertige Byte) an der niederen Speicher-Stelle abgespeichert; beim Big-Endian-Format wird dort
; das oberste Byte abgespeichert. Dies sind historisch gewachsende Formate; Little-Endian wird
; z.B. von den x86-Prozessoren von Intel und AMD verwendet, Big-Endian von PPC (IBMs Power-PC)
; und Motorolas 68000-er Serie (Amiga!).
;--------
!MOV eax,11223344h      ;Hex zur besseren bersicht 
!MOV [v_Var1],eax       ;nur fr die Anzeige
!BSWAP eax
!MOV [v_Var2],eax
MessageRequester("BSWAP", "Vor BSWAP :    " + Hex(Var1) + #LFCR$ + "Nach BSWAP : " + Hex(Var2))
;--------
; BSWAP ist "reversibel", d.h. es geht keine Information verloren; 2-mal hintereinander 
; ausgefhrt ergibt es wieder den Ausgangs-Wert.

; Zum Tauschen zweier Operanden gibt es XCHG (Exchange):
; Syntax: XCHG Operand1, Operand2. Beeinflusste Flags: Keine
; XCHG tauscht die Inhalte der beiden Operanden. Die Operanden mssen gleich gro sein und
; mindestens ein Operand mu ein Register sein:
;--------
!MOV eax,12345
!MOV [v_Var1],67890
!XCHG eax,[v_Var1]
!MOV [v_Var2],eax
MessageRequester("XCHG", "Nach XCHG  : Var1 = " + Str(Var1) + " und EAX = " + Str(Var2))
;--------

; Will man sich mal richtig am Flag-Register vergreifen, ist dies mit LAHF und SAHF mglich.
; LAHF (Load Status-Flags into AH-Register) und SAHF (Store AH-Register into Status-Flags)
; bentigen keinen Operanden, da beide von Hause aus mit dem AH-Register arbeiten. LAHF
; verndert keine Flags und bei SAHF hngt es vom Wert von Register AH ab. Manipulieren kann
; man mit SAHF die 8 unteren Bits des (E)Flag-Registers, also folgende Flags:
;       +----+----+----+----+----+----+----+----+
; Bit:  |Bit7|Bit6|Bit5|Bit4|Bit3|Bit2|Bit1|Bit0|
;       +----+----+----+----+----+----+----+----+
; Flag: | S  | Z  |    | A  |    | P  |    | C  |    leere Bit-Felder werden nicht benutzt!
;       +----+----+----+----+----+----+----+----+
; Wie man sieht, fllt z.B. das Overflow-Flag nicht darunter (ist Bit11).
; Im folgenden Beispiel lesen wir mal das A-Flag aus uns setzen P auf 1:
;--------
LAHF_String$ = "nicht "
!LAHF                   ;untere 8-Bit vom Flag-Register in AH schreiben
!TEST ah,00010000b      ;Bit4 testen
!JZ @f                  ;springe, wenn nicht gesetzt
LAHF_String$ = ""
!@@:
MessageRequester("LAHF", "Das Auxiliary-Flag war " + LAHF_String$ + "gesetzt!")

SAHF_String$ = "nicht "
!LAHF                   
!OR ah,00000100b        ;Bit2 setzen und die anderen Flags nicht verndern
!SAHF                   ;AH in Flag-Register schreiben
!LAHF                   ;zum Test wieder auslesen
!TEST ah,00000100b      ;Bit2 testen
!JZ @f                  ;springe, wenn nicht gesetzt
SAHF_String$ = ""
!@@:
MessageRequester("SAHF", "Das Parity-Flag war " + SAHF_String$ + "gesetzt!")
;--------

; Bevor es mit den doch recht umfangreichen String-Befehlen weitergeht noch was Einfaches:
; Syntax: NOP. Keine Operanden. Beeinflusste Flags: Keine
; NOP (No Operation) macht das, was sein Name schon sagt: Nichts! Aber er wird verwendet,
; um den Instruction-Pointer um Eins zu erhhen. Dies wird ntig, wenn ein nachfolgender
; Befehl (z.B. aus Geschwindigkeits-Grnden) auf eine durch 4 (auch 8 oder 16) teilbare
; (ohne Rest!) Speicher-Adresse liegen soll (mu). Viele Befehle werden vom Prozessor schneller
; eingelesen, wenn sie auf so einer Adresse liegen. Man kann dann auch mehrere NOP-Befehle
; hintereinander einstreuen; aber man sollte immer bedenken, das auch dieser Befehl CPU-Arbeitszeit
; verbrt. 
; Selbst NOPs im Code einzufgen wre Glcksspiel, deshalb gibt es die FAsm-Direktive ALIGN x.
; ALIGN (Alignment = Ausrichtung) ist kein Prozessor-Befehl! Der Assembler fgt in den Code
; die notwendige Anzahl NOPs ein, um damit auf die mit x angegebene Speicher-Ausrichtung zu
; kommen. Ein vorgefundenes Align 4 bedeutet fr den Assembler, soviel NOPs einzufgen, bis
; der nachfolgende Befehl auf einer durch 4 teilbaren Speicher-Adresse landet (also 0-3 NOPs).
; Grssere Werte fr x als 8 sind unter PB nicht ohne weiteres machbar.
;--------
!Align 4           ;hier fgt FAsm wenn erforderlich NOPs ein, damit der nachfolgende
                   ; MOV-Befehl auf einer restfrei durch 4 teilbaren Speicher-Adresse landet
!MOV [v_Var1],123456    

; der erzeugte ausfhrbare Code knnte so aussehen:
!NOP               ;zur Anzahl der NOPs s.o.! 
!NOP
!MOV [v_Var1],123456
;--------
; Werte fr x wie 1 oder 2 kann man sich schenken.

; Recht selten eingesetzt werden die beiden (identischen) Tabellen-Einlese-Befehle XLAT (Table
; Look-Up Translation) und XLATB. Wir verwenden hier XLATB, wobei das "B" gleich ausdrckt,
; das Byte-Operationen stattfinden.
; Syntax: XLATB. Keine Operanden. Beeinflusste Flags: Keine 
; XLATB benutzt zur Speicher-Adressierung das EBX-Register sowie den Wert vom Register AL.  
; ber diese Adresse wird in AL ein 8-Bit-Wert aus einer Tabelle, die nicht grsser als 256
; Bytes sein kann, eingelesen:
;--------
  DataSection
Tab_XLATB:              ;mit diesem Label kann FAsm nichts anfangen!
    Data.s  " ABCDEFGHIJKLMNOPQRSTUVWXYZ"   ;dieses Beispiel funktioniert nicht mit Unicode!
  EndDataSection

Var1 = ?Tab_XLATB       ;Var1 mit Start-Speicher-Adresse der Tabelle laden 
!MOV ebx,[v_Var1]       ;Start-Adresse des Buchstaben-Strings
!MOV al,13              ;die 13.Position soll ausgelesen werden; es gibt auch die nullte Pos.!      
!XLATB
!MOV [v_Var4],al        ;in AL steht der ausgelesene Wert 
MessageRequester("XLATB", "Der 13.Buchstabe des Alphabets ist das " + Chr(Var4))
;--------
; Das fhrende Leerzeichen im Buchstaben-String stellt den "nullten" Buchstaben dar!

; Das ermitteln der String-Speicher-Adresse kann auch mit dem Befehl LEA (Load Effective
; Address) erfolgen:
; Syntax: LEA Operand1, Operand2. Beeinflusste Flags: Keine
; Operand1 muss ein 16- oder 32-Bit-Register sein und Operand2 eine (Speicher-)Variable oder
; ein Label. PB-Labels kennt FAsm nicht!
; LEA ldt in Operand1 die (Offset-) Speicher-Adresse von Operand2. In 32-Bit-Betriebs-Systemen
; mu Operand1 natrlich ein 32-Bit-Register sein!
;--------
  DataSection
!Tab_XLATB:             ;dieses Label bleibt PB verborgen!
    Data.s  " ABCDEFGHIJKLMNOPQRSTUVWXYZ"   ;dieses Beispiel funktioniert nicht mit Unicode!
  EndDataSection

!LEA ebx,[Tab_XLATB]    ;Start-Adresse des Buchstaben-Strings ermitteln
!MOV al,23              ;die 23.Position soll ausgelesen werden; es gibt auch die nullte Pos.!      
!XLATB
!MOV [v_Var4],al        ;in AL steht der ausgelesene Wert 
MessageRequester("XLATB", "Der 23.Buchstabe des Alphabets ist das " + Chr(Var4))
;--------

; Wenn der Daten-String nur von FAsm benutzt wird, kann die DataSection von PB durch die FAsm-
; spezifische Data-Section ersetzt werden:
;-------- 
!LEA ebx,[Tab_XLATB1]   ;Start-Adresse des Buchstaben-Strings ermitteln
!MOV al,3               ;die 3.Position soll ausgelesen werden; es gibt auch die nullte Pos.!      
!XLATB
!MOV [v_Var4],al        ;in AL steht der ausgelesene Wert  
MessageRequester("XLATB", "Der 3.Buchstabe des Alphabets ist das " + Chr(Var4))

!JMP @f                 ;sonst wird versucht, die Tabelle als CPU-Code zu interpretieren!
!section '.data' Data readable writeable    ;dieser Teil kann gelesen UND beschrieben werden
!Tab_XLATB1 DB " ABCDEFGHIJKLMNOPQRSTUVWXYZ"     ;DB = Define Byte 
!@@:
;--------
; blicherweise setzt man die Data-Section an das Ende des Programmes und erspart sich so 
; die obigen Klimmzge. Man kann aber auch beliebig viele Data-Sections nach obigen Muster
; verwenden.   

; Da LEA ja den Adress-Offset berechnen mu, wird es auch gern mal eben zum "Rechnen missbraucht".
; Nachfolgender Code entspricht einer Multiplikation von EAX mit 5:
;--------
!MOV eax,15
!LEA eax,[eax+eax*4]
!MOV [v_Var1],eax  
MessageRequester("Multiplikation mit LEA ", "Dies entspricht einer Multiplikation mit 5 und das Ergebnis lautet: " + Str(Var1))
;--------
; Zu beachten ist, das hier keine Flags verndert werden, was aber durchaus auch mal erwnscht
; sein kann.

; So, jetzt kommen wir zu den eigentlichen String-Befehlen. Da interessieren uns 5 Gruppen:
; CMPS = Compare String; Vergleiche String
; LODS = Load String; Lade String
; MOVS = Move String; Kopiere String
; SCAS = Scan String; Scanne String
; STOS = Store String; Speichere String
; Jeden dieser Befehle gibt es in der allgemeinen Form wie oben sowie in je einer speziellen
; Ausfhrung fr Bytes, Words und Double-Words. Die allgemeine Form bentigt Operanden und
; gilt als veraltet, die spezielle Ausfhrung kommt ohne Operanden aus und sollte (auch lt.
; Intel) bevorzugt verwendet werden. Und daran halten wir uns dann.

; Zu Strings noch eine allgemeine Bemerkung: Strings sind nicht nur als Folgen von Buchstaben,
; Ziffern usw., also als Zeichenketten, zu verstehen, sondern ganz allgemein als mit Daten
; gefllte aufeinander folgende Speicher-Stellen. Wie diese Speicher-Bereiche zu interpretieren
; sind liegt ganz allein beim Programmierer! 

; Fr alle String-Befehle gilt: Die Adressierung erfolgt ber das ESI-Register als Quelle
; und das EDI-Register als Ziel. Ob die Adressen hoch- oder runter-gezhlt werden bestimmt
; das Direction-Flag: Ist es gelscht, wird hochgezhlt, ist es gesetzt, wird runtergezhlt.
; Zum Manipulieren des Direction-Flags gibt es die Befehle CLD (Clear Direction Flag) und
; STD (Set Direction Flag).
; Das Erhhen bzw. Verringern von ESI und EDI erfolgt automatisch nach jedem String-Befehl in
; der entsprechenden Grsse: Bei Bytes um jeweils 1, bei Words um 2 und bei DWords um 4.
; Der Programmierer ist aber dafr verantwortlich, vor dem Aufruf eines String-Befehls das
; Direction-Flag sowie ESI und EDI (beide je nach Befehl) richtig zu setzen.
 
; Jeder String-Befehl fhrt seine Aufgabe nur einmal aus; um mehrere Werte zu bearbeiten mu
; also so was wie eine Schleife ran. Dafr gibt es die Repeat-Befehle:
; Syntax: REP. Keine Operanden. Beeinflusste Flags: Keine
; Die String-Operation wird so lange ausgefhrt, wie ECX nicht Null ist.
; Syntax: REPE (=REPZ). Keine Operanden. Beeinflusste Flags: Keine
; Die String-Operation wird so lange ausgefhrt, wie ECX nicht Null oder das Zero-Flag gesetzt
; ist.
; Syntax: REPNE (=REPNZ). Keine Operanden. Beeinflusste Flags: Keine
; Die String-Operation wird so lange ausgefhrt, wie ECX nicht Null oder das Zero-Flag gelscht
; ist. 
; Nebenbei: Eigentlich sind die Repeat-Befehle gar keine "richtigen" CPU-Instruktionen, sondern
; sogenannte "Prfixe", die die Eigenschaften nachfolgender Instruktionen bestimmen.  
; Es bleibt brigens dem Programmierer berlassen, ob er die Repeat-Befehle oder eigene Abfragen
; einsetzt.

; Syntax: CMPSB. Keine Operanden. Beeinflusste Flags: C, P, A, Z, S, O.  
; Compare String Byte; Vergleiche String Byte
; Syntax: CMPSW. Keine Operanden. Beeinflusste Flags: C, P, A, Z, S, O.  
; Compare String Word; Vergleiche String Word 
; Syntax: CMPSD. Keine Operanden. Beeinflusste Flags: C, P, A, Z, S, O.  
; Compare String DWord; Vergleiche String DWord

; Nachfolgendes Beispiel berprft bei 2 gleichgrossen Strings, ob an einer Stelle eine 
; bereinstimmung vorhanden ist und bricht dann ab:
;--------
  DataSection           ;nehmen wir hier mal wieder
Tab_CMPSB1:             ;mit diesem Label kann FAsm nichts anfangen!
    Data.s  "123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"     ;Buchstaben und Ziffern wegen besserer 
Tab_CMPSB2:                                           ; berschaubarkeit; nicht fr Unicode!
    Data.s  "ZYXWVUTSRQPONMLKJIHGFEDCBA987654321"
  EndDataSection

Var1 = ?Tab_CMPSB1      ;Var1 mit Start-Speicher-Adresse des 1.Strings laden 
Var2 = ?Tab_CMPSB2      ;Var2 mit Start-Speicher-Adresse des 2.Strings laden 
Var8 = Var2 - Var1      ;wird hier der Einfachheit wegen als Stringlnge genommen 

!MOV esi,[v_Var1]       ;Start-Adresse des 1.Strings, hier als Quelle (Source) fungierend
!MOV edi,[v_Var2]       ;Start-Adresse des 2.Strings, hier als Ziel (Destination) fungierend
!MOV ecx,[v_Var8]       ;Lnge der Strings, darber kann man sich vorher auch andere Gedanken
                        ; machen
!MOV [v_Var7],0         ;Erfolgs-Merker, s.u.  
!CLD                    ;Direction-Flag lschen, also die Strings speichermssig aufwrts einlesen 
!REPNZ                  ;maximal simples Suchen bis ECX null ist oder eine bereinstimmung
                        ; gefunden wurde, also dann das Zero-Flag gesetzt ist
!CMPSB                  ;vergleicht byte-weise auf bereinstimmung; die Strings werden nicht
                        ; verndert 
!JECXZ @f               ;dieser Befehl (s.o.) bietet sich hier an, um den Erfolg zu testen: 
                        ; Wurde ECX auf Null runter gezhlt, ist keine bereinstimmumg gefunden
                        ; worden. Aber Vorsicht: Mit ihm sind nur Sprunge von max. 127 Bytes
                        ; vorwrts mglich! ber einen etwas lngeren Messagerequester kommt er
                        ; nicht drber, deshalb hier der Umweg ber einen Merker (Var7)    
!MOV [v_Var7],1         ;Erfolg vermerken
!MOV [v_Var1],ecx       ;den Restwert von ECX nach Var1 kopieren 
!MOV al,[esi-1]         ;den Wert, bei dem Gleichheit festgestellt wurde, einlesen. ESI-1, weil
                        ; CMPSB das Register ESI schon wieder um 1 erhht hatte (EDI auch)
!MOV [v_Var4],al
!@@:                    ;Sprungziel von JECXZ; musste innerhalb der nchsten 127 Byte liegen  
!CMP [v_Var7],0         ;bereinstimmung gefunden?
!JE @f                  ;nein
MessageRequester("CMPSB", "bereinstimmung besteht an " + Str(Var8 - Var1) + ".Stelle und es ist das " + Chr(Var4))
!JMP l_erfolg
!@@:
MessageRequester("CMPSB", "Keine bereinstimmung gefunden!")    ;einfach mal ein "I" ndern!
Erfolg:
;--------
; Die Klimmzge mit dem Merker usw. wurden hier nur wegen der Anzeige mit dem Messagerequester
; gemacht! In "richtigem" ASM-Code reichen 127 Byte als Sprung-Bereich blicherweise aus, JECXZ
; und JCXZ sind ja sowieso Sonderflle, was die Sprungweite betrifft.

; Die obigen Klimmzge knnte man z.B. auch dadurch umgehen, den Messagerequester in eine Sub-
; Routine zu packen und per CALL aufrufen. Aber es geht natrlich auch komplett in Assembler.
; Gleich vorweg: Hier wurde ein wenig geschummelt, um es nicht zu kompliziert zu machen: Die 
; "Fundstelle" wurde in den einstelligen Bereich gelegt ("I" an 8.Stelle), um Umrechnungen in
; mehrstellige Zahlen zu umgehen.
;--------
!extrn _MessageBoxA@16  ;damit ldt FAsm die fr die MessageBox bentigte Bibliothek
                        ; _MessageBoxW ist die Version fr Unicode 
!LEA esi,[Tab_CMPSB1]   ;Start-Adresse des 1.Strings, hier als Quelle (Source) fungierend
!LEA edi,[Tab_CMPSB2]   ;Start-Adresse des 2.Strings, hier als Ziel (Destination) fungierend
!MOV ecx,edi            ;fr Ermittlung der Stringlnge von Tab_CMPSB1 
!SUB ecx,esi
!MOV edx,ecx            ;sichern fr Stelle berechnen

!CLD                    ;Direction-Flag lschen, also die Strings speichermssig aufwrts einlesen 
!REPNZ                  ;maximal simples Suchen bis ECX null ist oder eine bereinstimmung
                        ; gefunden wurde, also dann das Zero-Flag gesetzt ist
!CMPSB                  ;vergleicht byte-weise auf bereinstimmung; die Strings werden nicht
                        ; verndert 
!JECXZ @f               ;dieser Befehl (s.o.) bietet sich hier an, um den Erfolg zu testen: 
                        ; Wurde ECX auf Null runter gezhlt, ist keine bereinstimmumg gefunden
                        ; worden. Aber Vorsicht: Mit ihm sind nur Sprunge von max. 127 Bytes
                        ; vorwrts mglich!    
!SUB edx,ecx            ;Stelle berechnen         
!ADD dl,48              ;ergibt ASCII-Wert
!MOV [Stelle],dl        ;hier nur ein Byte (Stelle 8) 
!MOV al,[esi-1]         ;den Wert, bei dem Gleichheit festgestellt wurde, einlesen. ESI-1, weil
                        ; CMPSB das Register ESI schon wieder um 1 erhht hatte (EDI auch)
!MOV [Zeichen],al
;-- beim Aufruf von _MessageBoxA werden 4 Parameter auf dem Stack erwartet
!PUSH 0                 ;bestimmt die MessageBox-Buttons, 0 = nur OK
!PUSH Titel             ;Titel der MessageBox
!PUSH Message1          ;Text in der MessageBox
!PUSH 0                 ;Handle, hier erstmal uninteressant; Hauptsache 0
!CALL _MessageBoxA@16   ;das @16 bedeutet, die 16 Bytes der 4 Parameter (a 4 Bytes) werden 
                        ; automatisch vom Stack getilgt (ESP um 16 erhht)
;--
!JMP Erfolg_gehabt
!@@:                    ;Sprungziel von JECXZ; musste innerhalb der nchsten 127 Byte liegen  
;--
!PUSH 0
!PUSH Titel
!PUSH Message2
!PUSH 0
!CALL _MessageBoxA@16
;--
!Erfolg_gehabt:
!JMP @f                 ;um den Daten-Teil zu berspringen  
;---
!section '.data' Data readable writeable
!Tab_CMPSB1  DB "1234567I89ABCDEFGHJKLMNOPQRSTUVWXYZ"
!Tab_CMPSB2  DB "ZYXWVUTISRQPONMLKJHGFEDCBA987654321"

!Titel    DB "CMPSB in reinem Assembler",0       ;0 kennzeichnet das Stringende ("Zero-Byte")
!Message1 DB "bereinstimmung besteht an "
!Stelle   DB 1                                   ;Platzhalter fr Stelle, hier beliebiger Wert,
                                                 ; aber nicht 0, das wrde als Stringende von 
                                                 ; Message1 interpretiert werden!               
!         DB ".Stelle und es ist das " 
!Zeichen  DB 1,0                                 ;Platzhalter fr Zeichen, 0 kennzeichnet das 
                                                 ; Stringende von Message1
!Message2 DB "Keine bereinstimmung gefunden!",0
;---
!@@:                    ;von Daten-Teil berspringen
;--------
; Zum Stringende kennzeichnen: Sollte so wie oben gehandhabt werden, auch wenn Tests dies nicht
; unbedingt fr ntig erscheinen lassen!
; Ich hoffe, obiges Beispiel ist nicht zu unbersichtlich. Wie schon einmal bemerkt, landet die
; Data-Section blicherweise am Ende des Programmes und eine Anweisung wie _extrn am Beginn des
; Programmes.
; Obiges Beispiel mit "Nur-Assembler-Code" soll sowieso eine Ausnahme bleiben, denn wir wollen
; ja die Vorzge beider Sprachen ausnutzen, und z.B. die Bildschirm-Ausgabe ist mit PB nun mal
; einfacher und bequemer zu realisieren.

; Beispiele fr CMPSW und CMPSD schenken wir uns hier; es luft analog wie oben. Es ist lediglich
; zu beachten, das ESI/EDI in Zweier- bzw. Vierer-Schritten verndert werden und ECX als Zhler
; fr den Repeat-Befehl dementsprechend anzupassen ist.

; Kommen wir zu LODSx und STOSx:
; Syntax: LODSB. Keine Operanden. Beeinflusste Flags: Keine  
; Load String Byte; Lade String Byte
; Syntax: LODSW. Keine Operanden. Beeinflusste Flags: Keine  
; Load String Word; Lade String Word 
; Syntax: LODSD. Keine Operanden. Beeinflusste Flags: Keine  
; Load String DWord; Lade String DWord 
; Adressiert wird ber das ESI-Register, welches dann um 1 (Byte), 2 (Word) bzw. 4 (DWord)
; verndert wird. Die Richtung bestimmt wieder das Direction-Flag.
; Der eingelesene Wert wird in das AL- (Byte), AX- (Word) bzw. EAX- (DWord) Register geladen.

; Syntax: STOSB. Keine Operanden. Beeinflusste Flags: Keine  
; Store String Byte; Speichere String Byte
; Syntax: STOSW. Keine Operanden. Beeinflusste Flags: Keine  
; Store String Word; Speichere String Word 
; Syntax: STOSD. Keine Operanden. Beeinflusste Flags: Keine  
; Store String DWord; Speichere String DWord 
; Adressiert wird ber das EDI-Register, welches dann um 1 (Byte), 2 (Word) bzw. 4 (DWord)
; verndert wird. Die Richtung bestimmt wieder das Direction-Flag.
; Gespeichert wird der Wert aus dem AL- (Byte), AX- (Word) bzw. EAX- (DWord) Register.

; Unser Beispiel ldt Words aus der DataSection, inkrementiert den Wert und speichert ihn an
; die selbe Stelle zurck. Abbruch, wenn ein Wert kleiner Null eingelesen wurde.
;--------
  DataSection
Tab_LOSTSW:
    Data.w  1,2,3,4,5,6,7,8,9,10,11,12,13,-1
  EndDataSection

Var1 = ?Tab_LOSTSW      ;Var1 mit Start-Speicher-Adresse des Strings laden 

!MOV esi,[v_Var1]       ;Start-Adresse des Strings, hier als Quelle (Source) fungierend
!MOV edi,esi            ;frs zurckspeichern
!CLD                    ;Direction-Flag lschen, also Speicher vorwrts abklappern
!@@:
!LODSW                  ;Wert in AX einlesen; ESI+2            
!TEST ax,1000000000000000b   ;Test auf negativ (zur Erinnerung: MSB = Vorzeichen)
!JS @f                  ;MSB ist gesetzt, also negativ. Sprung zum Ende
!INC ax                 ;AX inkrementieren         
!STOSW                  ;AX an die selbe Stelle zurckschreiben; EDI+2
!JMP @b                 ;nchsten Wert einlesen
!@@:
;- um hier nicht viel Aufwand zu treiben, wird einfach nur der 1.Wert wieder ausgelesen 
!MOV ebx,[v_Var1]       ;wieder (noch) Start-Adresse
!MOV dx,[ebx]           ;1.Wert einlesen
!MOV [v_Var5],dx        ;zur Anzeige in Word-Variable kopieren 
MessageRequester("LODSW / STOSW", "Der Wert an der 1.Stelle betrgt jetzt " + Str(Var5)) 
;--------
; Schlicht und einfach; sollte klar sein.

; Nachfolgendes Beispiel berprft, wie oft der Buchstabe "K" im Vergleichsstring enthalten ist:
;--------
LODSB1$ = "ABCDEFKGHIJKLMNOPQRSTUKVWXYZ"  ;jetzt mal ohne DataSection. Nicht fr Unicode!
Var1 = @LODSB1$         ;Start-Speicher-Adresse vom String

!MOV esi,[v_Var1]       ;LODSx benutzt ESI zur Adressierung
!MOV [v_Var2],0         ;Zhler fr Anzahl gefundenes Zeichen
!CLD                    ;Direction-Flag lschen; Speicher aufwrts zhlen
!@@:
!LODSB                  ;1 String-Byte in AL einlesen
!OR al,al               ;hier Test auf das "Zero-Byte", welches das Ende eines Strings signalisiert 
!JE @f                  ;Stringende erreicht
!CMP al,"K"
!JNE @b                 ;nicht gefunden, weiter ohne Zhler zu erhhen
!INC [v_Var2]           ;Zhler erhhen
!JMP @b                 ;und weitersuchen 
!@@:
MessageRequester("LODSB", "Der Buchstabe 'K' ist " + Str(Var2) + "-mal im String enthalten") 
;--------

; Weiter mit MOVSX:
; Syntax: MOVSB. Keine Operanden. Beeinflusste Flags: Keine  
; Move Data from String to String Byte; Kopiere Daten von String zu String Byte
; Syntax: MOVSW. Keine Operanden. Beeinflusste Flags: Keine  
; Move Data from String to String Word; Kopiere Daten von String zu String Word 
; Syntax: MOVSD. Keine Operanden. Beeinflusste Flags: Keine  
; Move Data from String to String DWord; Kopiere Daten von String zu String DWord 
; Adressiert wird ber das ESI-Register fr den Quell-String und ber das EDI-Register fr den
; Ziel-String; die Register werden dann um 1 (Byte), 2 (Word) bzw. 4 (DWord) verndert. Die
; Richtung bestimmt wieder das Direction-Flag.

; Nachfolgendes Beispiel kopiert 10 Long-Werte aus der DataSection in einen Ziel-String:
;--------
  DataSection
Tab_MOVSD1:
    Data.l  1,2,3,4,5,6,7,8,9,10  ;Quell-String
Tab_MOVSD2:
    Data.l  0,0,0,0,0,0,0,0,0,0   ;Ziel-String
  EndDataSection

Var1 = ?Tab_MOVSD1      ;Var1 mit Start-Speicher-Adresse des Quell-Strings laden 
Var2 = ?Tab_MOVSD2      ;Var2 mit Start-Speicher-Adresse des Ziel-Strings laden 

!MOV esi,[v_Var1]       ;Start-Adresse des Quell-Strings
!MOV edi,[v_Var2]       ;Start-Adresse des Ziel-Strings
!MOV ecx,10             ;mal ganz einfach
!CLD                    ;Direction-Flag lschen, also Speicher vorwrts abklappern
!REP                    ;einfach nur ECX runterzhlen
!MOVSD                  ;von DataSection String1 in String2 kopieren            
;- um auch hier nicht viel Aufwand zu treiben, wird einfach nur der 1.Wert ausgelesen 
!MOV ebx,[v_Var2]       ;noch Start-Adresse vom Ziel-String
!MOV edx,[ebx]          ;1.Wert einlesen
!MOV [v_Var8],edx       ;zur Anzeige in DWord-Variable kopieren 
MessageRequester("MOVSD", "Der 1.Wert im vorher leeren Ziel-String ist " + Str(Var8)) 
;--------
; Nochmal: Den Begriff String nicht zu eng sehen!
; Mit MOVSD kann man schnell grssere Speicher-Bereiche umkopieren.

; Fehlt noch SCASx:
; Syntax: SCASB. Keine Operanden. Beeinflusste Flags: C, P, A, Z, S, O.  
; Scan String Byte; Scanne String Byte
; Syntax: SCASW. Keine Operanden. Beeinflusste Flags: C, P, A, Z, S, O.  
; Scan String Word; Scanne String Word 
; Syntax: SCASD. Keine Operanden. Beeinflusste Flags: C, P, A, Z, S, O.  
; Scan String DWord; Scanne String DWord 
; Adressiert wird ber das EDI-Register, welches dann um 1 (Byte), 2 (Word) bzw. 4 (DWord)
; verndert wird. Die Richtung bestimmt wieder das Direction-Flag.
; Das Scannen ist praktisch ein Vergleichen von Register AL (Byte), AX (Word) bzw. EAX (DWord)
; mit dem eingelesenen Wert. Bei Gleichheit werden die wie bei den "gewhnlichen" Vergleichs-
; Befehlen beteiligten Flags gesetzt, wobei hier hauptschlich das Zero-Flag (gesetzt bei
; Gleichheit, sonst gelscht) interessiert.  

; Nachfolgendes Beispiel ermittelt von einem String die Lnge:
;--------
SCASB1$ = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"      ;Nicht fr Unicode!
Var1 = @SCASB1$         ;Start-Speicher-Adresse vom String

!MOV edi,[v_Var1]       ;SCASx benutzt EDI zur Adressierung
!MOV [v_Var2],0         ;Zhler fr Anzahl gefundenes Zeichen
!XOR al,al              ;"Zero-Byte", welches das Ende eines Strings signalisiert
!CLD                    ;Direction-Flag lschen; Speicher aufwrts zhlen
!@@:
!SCASB                  ;testet hier also auf Null (nicht die Ziffer Null, sondern "richtig" Null!)
!JE @f                  ;Stringende erreicht
!INC [v_Var2]           ;Zhler erhhen
!JMP @b                 ;und weitersuchen
!@@:
MessageRequester("SCASB", "Der String ist " + Str(Var2) + " Zeichen lang") 
;--------

; Kommen wir nun zu den BCDs, den Binary Coded Decimals. Man glaubt es kaum, aber ein (x86-)
; Prozessor kann sogar mit Dezimalzahlen umgehen; wenn auch sehr rudimentr. Selbst die FPU
; kann Ergebnisse im BCD-Format liefern. Was sind nun diese BCDs? Ganz einfach: Der Inhalt
; eines Bytes wird als eine Dezimal-ZIFFER betrachtet! Diese Verfahrensweise wird als ungepackte
; BCD bezeichnet. Da Dezimal-Ziffern ja von 0-9 gehen, ein Nibble (4 Bits eines Bytes) aber max.
; den Wert 15 dezimal (0F hex) darstellen kann, gibt es noch die Platzspar-Variante gepackte
; BCD: Jedes Nibble reprsentiert eine Dezimal-Ziffer; ein Byte also 2 Ziffern. Die Bits eines
; Bytes bzw. Nibbles werden also als binre Codierung einer Dezimal-Ziffer betrachtet; daher der
; Name.
; Fr BCDs gibt es 6 Prozessor-Befehle: AAA, AAD, AAM, AAS, DAA und DAS. Alle diese Befehle
; fhren Korrektur-Arbeiten durch, benutzt wird immer das AL- bzw. AX-Register. 
; Syntax: AAA. Keine Operanden. Beeinflusste Flags: C, P, A, Z, S, O. Fr ungepackte BCDs.
; AAA = ASCII Adjust After Addition. Das obere Nibble von AL wird gelscht und es wird berprft,
; ob der Wert des unteren Nibbles von AL grsser als 9 ist; wenn ja, wird 6 hinzugezhlt und
; AH wird inkrementiert.     
;--------
Global Var9.b
!MOV ah,4
!MOV al,44              ;2ch
!MOV [v_Var4],al
!AAA
!MOV [v_Var7],al
!MOV [v_Var9],ah
MessageRequester("AAA", "Register AL vorher:    " + RSet(Bin(Var4,#PB_Byte),8,"0") + #LFCR$ + "Register AL nachher: " + RSet(Bin(Var7,#PB_Byte),8,"0") + #LFCR$ + "AH jetzt :  " + Str(Var9))  
;--------
; Das untere Nibble von AL hatte den Wert 12, nach AAA 2 und AH wurde um 1 erhht. Wre AH Null
; gewesen, wrde in AH 1 und in AL 2 stehen, also in AX die ungepackte BCD-Zahl 12. Nach einer
; BCD-Addition kann man also mit AAA das Ergebnis wieder BCD-konform korrigieren. Das ASCII im
; Namen kommt daher, da durch jeweilige Addition von 30hex in AH und AL die entsprechenden 
; ASCII-Ziffern stehen, die so direkt z.B. zu einer Bildschirm-Ausgabe gesendet werden knnten
; (in unseren Beispielen erledigt das der Str()-Befehl).
; Eine 2-stellige Addition geht dann so:  
;--------
!MOV ah,5               ;Zehner
!MOV al,6               ;Einer
!MOV bh,2               ;Zehner    
!MOV bl,8               ;Einer
!ADD ax,bx              ;56+28 im BCD-Format, sollte 84 ergeben
!AAA
!MOV [v_Var7],al        ;Einer Ergebnis
!MOV [v_Var9],ah        ;Zehner Ergebnis
MessageRequester("AAA", "56 + 28 ergibt:" +  #LFCR$ + "Zehner: " + Str(Var9) + #LFCR$ + "Einer:    " + Str(Var7))  
;--------
; Mal ohne AAA testen!
; Zu beachten ist generell der begrenzte Wertebereich von 00 bis maximal 99. 

; Syntax: AAS. Keine Operanden. Beeinflusste Flags: C, P, A, Z, S, O. Fr ungepackte BCDs.
; AAS = ASCII Adjust AL After Subtraction. Das obere Nibble von AL wird gelscht und es wird
; berprft ob der Wert des unteren Nibbles von AL grsser als 9 ist; wenn ja, wird 6
; abgezogen und AH wird dekrementiert. AAS ist nur fr einstellige BCDs gedacht! Fr eine
; zweistellige Subtraktion kann man das Carry-Flag (gesetzt, wenn Korrektur vorgenommen wurde)
; auswerten und AH wieder korrigieren.    
; Eine 2-stellige Subtraktion geht dann so:  
;--------
!MOV ah,8               ;Zehner
!MOV al,4               ;Einer
!MOV bh,3               ;Zehner    
!MOV bl,8               ;Einer
!SUB ax,bx              ;84-38 im BCD-Format, sollte 46 ergeben
!AAS                    ;dekrementiert bei Korrektur auch AH
!JNC @f                 ;wurde korrigiert (dann Carry gesetzt)?
!INC ah                 ;ja, AH wieder inkrementieren
!@@:
!MOV [v_Var7],al        ;Einer Ergebnis
!MOV [v_Var9],ah        ;Zehner Ergebnis
MessageRequester("AAS", "84 - 38 ergibt:" +  #LFCR$ + "Zehner: " + Str(Var9) + #LFCR$ + "Einer:    " + Str(Var7))  
;--------

; Die beiden obigen Befehle sind fr ungepackte BCDs. Fr die Addition und Subtraktion (und
; nur fr diese!) gibt es auch Befehle fr gepackte BCDs: DAA und DAS.
; Syntax: DAA. Keine Operanden. Beeinflusste Flags: C, P, A, Z, S, O. Fr gepackte BCDs.
; DAA = Decimal Adjust AL After Addition. Es wird nur das AL-Register verwendet!
;--------
!MOV al,49h             ;Hex zur besseren Anzeige, da hier jedes Nibble eine Ziffer reprsentiert
!MOV bl,35h             ; 35h z.B. ist also als 35 dezimal zu verstehen! 
!ADD al,bl
!DAA                    ;korrigiert das Ergebnis von 7E zu 84
!MOV [v_Var7],al
MessageRequester("DAA", "49 + 35 ergibt:" +  #LFCR$ + "Ergebnis: " + Hex(Var7 & $ff))
;--------

; Syntax: DAS. Keine Operanden. Beeinflusste Flags: C, P, A, Z, S, O. Fr gepackte BCDs.
; DAS = Decimal Adjust AL After Subtraction. Es wird nur das AL-Register verwendet!
;--------
!MOV al,62h             ;ist praktische eine Dezimal-Zahl!
!MOV bl,35h 
!SUB al,bl
!DAS                    ;korrigiert das Ergebnis von 2D zu 27
!MOV [v_Var7],al
MessageRequester("DAS", "62 - 35 ergibt:" +  #LFCR$ + "Ergebnis: " + Hex(Var7 & $ff))
;--------

; Syntax: AAM. Keine Operanden. Beeinflusste Flags: C, P, A, Z, S, O. Fr ungepackte BCDs.
; AAM = ASCII Adjust AX After Multiply. Korrigiert das Ergebnis einer Multiplikation von zwei
; Einer-BCDs. AL wird durch 10 geteilt und der Ganzteil-Wert wird in AH kopiert, in AL landet
; der Rest (Modulo).
;--------
!MOV al,9
!MOV bl,7 
!MUL bl                 ;9*7
!AAM
!MOV [v_Var7],al
!MOV [v_Var9],ah
MessageRequester("AAM", "9 * 7 ergibt:" +  #LFCR$ + "Zehner: " + Str(Var9) + #LFCR$ + "Einer:    " + Str(Var7)) 
;--------

; Syntax: AAD. Keine Operanden. Beeinflusste Flags: C, P, A, Z, S, O. Fr ungepackte BCDs.
; AAD = ASCII Adjust AX Before Division. Hier wird man stutzig: Before? Die anderen Korrektur-
; Befehle wurden nach einer mathematischen Operation durchgefhrt und jetzt vorher? Ist aber
; wirklich so: Dividiert werden knnen maximal zweistellige BCD-Zahlen und fr den DIV-Befehl
; muss das so aufbereitet werden, das in AX der Dividend in hexadezimaler Form vorliegt. Dazu
; wird AH mit 10 multipliziert und zu AL zuaddiert, danach wird AH auf Null gesetzt. Die
; nachfolgende Division mu dann in der Form "DIV 8-Bit-Register" erfolgen, also praktisch
; DIV AX,8-Bit-Register. 
;--------
!MOV ah,4               ;Zehner
!MOV al,8               ;Einer vom Dividenden
!MOV bl,6               ;Divisor 
!AAD                    ;BCD in Hex umwandeln
!DIV bl                 ;48/6
!MOV [v_Var7],al
!MOV [v_Var9],ah
MessageRequester("AAD", "48 / 6 ergibt:" +  #LFCR$ + "Zehner: " + Str(Var9) + #LFCR$ + "Einer:    " + Str(Var7)) 
; -------

; Ist zu erwarten, da das Ergebnis im BCD-Sinne zweistellig sein kann, ist nach der Division
; natrlich eine weitere Korrektur ntig. Dazu dient hier AAM; also den beschriebenen Einsatz-
; Zweck (hier: Nach einer Multiplikation) nicht zu eng sehen!
;--------
!MOV ah,7               ;Zehner
!MOV al,8               ;Einer vom Dividenden
!MOV bl,6               ;Divisor 
!AAD                    ;wie oben
!DIV bl                 ;78/6
!AAM                    ;korrigiert jetzt AL
!MOV [v_Var7],al
!MOV [v_Var9],ah
MessageRequester("AAD", "78 / 6 ergibt:" +  #LFCR$ + "Zehner: " + Str(Var9) + #LFCR$ + "Einer:    " + Str(Var7)) 
; -------

; Der obige Trick mit AAM lsst sich generell so verwenden, das eine Hex-Zahl in AX, die nicht
; grsser als 99 dezimal sein darf, in zwei BCD-Ziffern in AH und AL zerlegt werden kann:
;--------
!MOV ax,56
!AAM
!MOV [v_Var7],al
!MOV [v_Var9],ah
MessageRequester("AAM", "56 in AX ergibt:" +  #LFCR$ + "Zehner: " + Str(Var9) + #LFCR$ + "Einer:    " + Str(Var7)) 
; ------- 
; Die 56 in AX steht ja nur im AL-Register, mit AAM wird sie in AH (Zehner) und AL (Einer)
; aufgeteilt.

; Generell gilt: Fr die Multiplikation und Division sind die vorzeichenlosen Versionen, also
; MUL und DIV, zu verwenden!

; Nachfolgendes Beispiel berechnet das Produkt zweier grossen Zahlen, die als String vorliegen.
; Verwendet werden ungepackte BCDs, korrigiert wird mit AAA und AAM:
;--------
Global pF1.l
Global pF2.l
Global F1L.l
Global F2L.l
Global pP.l
Global LP.l
Global ESICH.l
Global BSICH.l
Global LKLEIN.l
Global F1$ = "5555555556666666666777777777777777778765434567654"     ;Kein Unicode! Sonst anpassen!
Global F2$ = "123456789098765432198765432123456789"

F1L = Len(F1$)
F2L = Len(F2$)

If F1L < F2L                 ;Multiplikator (1.Faktor) muss der lngere String sein
   pF1 = @F2$                ;Pointer auf die Strings
   pF2 = @F1$ 
   Swap F1L,F2L
 Else
   pF1 = @F1$
   pF2 = @F2$
EndIf

LP = F1L + F2L               ;max. Lnge des Produkts
LKLEIN = F2L

Produkt$ = Space(LP)
pP = @Produkt$ 

;bei Verwendung z.B. in Prozeduren sind natrlich die Register EBX, EDI und ESI zu sichern und wieder herzustellen
!MOV esi,[v_pF1]             ;Pointer auf String1 (Speicher-Anfang)
!ADD esi,[v_F1L]             ;Lnge von String1 dazu addieren
!DEC esi                     ;zeigt so auf die letzte Ziffer von String1
!MOV [v_ESICH],esi

!MOV edi,[v_pF2]             ;Pointer auf String2
!ADD edi,[v_F2L]             ;Lnge von String2 dazu addieren
!DEC edi                     ;zeigt so auf die letzte Ziffer von String2

!MOV ebx,[v_pP]              ;Pointer auf zu berechnendes Produkt
!ADD ebx,[v_LP]              ;(max. zu erwartende) Lnge von Produkt dazu addieren
!dec ebx
!MOV [v_BSICH],ebx           ;sichern
                   
!XOR dh,dh                   ;DH=0, sonst Fehler mglich bei "ADD al,dh" (DH ist bertrag)
!MOV ecx,[v_F1L]

!MOV dl,[edi]                ;letzte (rechte) Ziffer von String2 einlesen
!SUB dl,30h                  ;ASCII-Ziffer (30h bis 39h) des Strings in Byte-Wert 0 bis 9 umwandeln
;--------
;in dieser Schleife wird jeweils eine Ziffer von String2 mit allen Ziffern von String1 multipliziert,
; und zwar werden die String1-Ziffern nacheinander von rechts nach links abgearbeitet
!@@:
!MOV al,[esi]                ;Ziffer von String1 einlesen
!SUB al,30h                  ;ASCII-Ziffer (30h bis 39h) des Strings in Byte-Wert 0 bis 9 umwandeln

!MUL dl                      ;AL*DL, also Ziffer von String1 * Ziffer von String2; da max. 9*9 mglich wird AH Null
!AAM                         ;Korrektur nach einer Multiplikation              
!ADD al,dh                   ;evtl. vorhandenen bertrag addieren 
!AAA                         ;Korrektur nach einer Addition
!ADD al,[ebx]                ;bereits vorhandener Wert von Produkt wird dazu addiert
!AAA                         ;in erster Runde siebt dies auch Space aus! Oberes Nibble von AL wird gelscht
!MOV [ebx],al
!MOV dh,ah                   ;(eventuellen) bertrag sichern 

!DEC esi                     ;zeigt auf nchste Ziffer (von rechts nach links) von String1
!DEC ebx                     ;zeigt auf nchste Ziffer (von rechts nach links) von Produkt
!DEC ecx                     ;noch abzuarbeitende Ziffern von String1 minus 1
!JNZ @b                      ;zurck, wenn noch Ziffer(n) von String1 da
;--------
!ADD [ebx],ah                ;Rest (bertrag von oben) zur Produkt-Stelle dazu addieren

!DEC [v_LKLEIN]              ;Test auf Ende, also keine Ziffer vom kurzen String (String2) mehr vorhanden
!JZ @f

!XOR dh,dh
!MOV ecx,[v_F1L]
!DEC edi                     
!MOV dl,[edi]                ;neue Ziffer aus 2.String
!SUB dl,30h
!MOV esi,[v_ESICH]
!MOV ebx,[v_BSICH]           ;Pointer in Produkt
!DEC ebx                     ;minus 1
!MOV [v_BSICH],ebx           ;Pointer in Produkt eine Stelle nach links
!JMP @b
;--------
!@@:
!AND byte[ebx],0Fh           ;Space ausblenden

!OR ah,ah                    ;Test auf fhrende Null
!JNE @f                      ;keine
!MOV byte[ebx],20h           ;Null mit Leerzeichen berschreiben
!INC ebx                     ;nachfolgende Umwandlung in ASCII-Zeichen kann eine Stelle weiter beginnen
!@@:
;Umwandlung des Produkts in ASCII-Zeichen
!MOV ecx,[v_pP]
!ADD ecx,[v_LP]
!SUB ecx,ebx
!@@:
!ADD byte[ebx],30h
!INC ebx
!DEC ecx
!JNZ @b

MessageRequester("BCD-Multiplikation", F1$ + " * " + F2$ + " = " + Trim(Produkt$))

;-------
; In der letzten Code-Zeile "MOV byte[ebx],20h" ist die Angabe "byte" zwingend notwendig, da der
; Compiler nicht wei, ob an der Speicher-Adresse ein Byte, ein Word oder sonstwas geschrieben
; werden soll. Der Wert "20h" knnte ja auch als ein DWord gemeint sein. Hier soll aber nur ein
; Leerzeichen geschrieben werden, daher byte.

;=========================================================================================
; So, das waren die wichtigen (und vielleicht auch einige unwichtige) Prozessor-Befehle.
; Wichtig und unwichtig ist natrlich relativ; der Programmierer setzt das ein, was fr sein
; Programm erforderlich ist.

; Jetzt wollen wir mal sehen, wie die Win32-API-Funktionen mit FAsm zu nutzen sind. API steht
; fr "Application Programming Interface" und ist in unserem Sinne als Bibliothek fr diverse
; (System-)Funktionen eines 32-Bit-Windows-Betriebssystems zu verstehen. 
; PB selbst verwendet nicht nur im "Hintergrund" die Funktionen der Win32-API, sondern man kann
; sie selbst auch ganz "offen" nutzen, ohne irgendwelche Vorbereitungen machen zu mssen. Als
; Beispiel nehmen wir mal die BEEP-Funktion, mit der man dem PC-Speaker Tne entlocken kann. PB
; bietet diesen Befehl nicht direkt an (d.h. er steht nicht in der PB-Befehls-bersicht), aber
; so kann man ihn anwenden (evtl. nicht aufm Laptop!):
Beep_(1500, 500)
; Die Ton-Frequenz betrgt 1500 Hz, die Ton-Dauer 500ms.  
; Am Unterstrich nach dem eigentlichen Befehl erkennt man in PB den Aufruf einer API-Funktion.
; Mit FASm sieht das Ganze so aus:
;--------
;!extrn _Beep@8          ;so wird FAsm mitgeteilt, das diese Funktion (spter) verwendet wird;
                        ; man mge sich doch darum kmmern. Diese Deklaration erfolgt 
                        ; blicherweise am Anfang eines Programmes 
!PUSH dword 500         ;Ton-Dauer
!PUSH dword 2000        ;Ton-Frequenz
!CALL _Beep@8           ;Aufruf der oben angemeldeten Beep-Funktion

MessageRequester("Akustik-Test", "War das nicht schn?" + #LFCR$ + "Aber jetzt erst weiterlesen, dann anklicken!")

;--------
; Wer noch nicht ganz Disco-geschdigt ist, sollte jetzt 2 kurze unterschiedliche Tne gehrt
; haben. Aber warum wurde die 1.Zeile auskommentiert? Einfach mal "gltig" machen und die
; Fehler-Meldung lesen! Dann sollte klar werden, da das 1.Beep genau die gleiche Funktion nutzt
; und die Beep-Funktion schon selbst deklariert hat. Einfach mal oben das Beep_() auskommentieren!
; Und was bedeutet das "@8"? Die Beep-Funktion erwartet 2 DWord-Parameter (also 8 Bytes) auf dem
; Stack; deshalb die beiden PUSH-Befehle vor dem Call. Und das "@8" signalisiert dem Programmierer,
; das er sich nicht um die Stack-Bereinigung nach dem Aufruf der Funktion kmmern mu, denn
; eigentlich haben ja nach 2 PUSHs auch 2 POPs zu folgen. Die Funktion korrigiert selbst nach
; ihrer Beendigung den Stack-Pointer; hier also plus 8. So ganz nebenbei kann man so auch die
; Anzahl der erwarteten (bentigten) Funktions-Parameter erkennen.
; Generell gilt: Die Parameter-bergabe an API-Funktionen erfolgt immer ber den Stack! Das
; dabei die richtige Reihenfolge wichtig ist, sollte jedem klar sein. Aber was ist, wenn zur
; Beschreibung einer API-Funktion sowas steht wie "Funktionsname(Parameter1, Parameter2)"? Wie
; ist die Reihenfolge im Programm? Bei 2 Parametern kann man ja noch ausprobieren, ab 5 wirds
; lstig. Also: Parameter1 ist immer der, der dem Funktions-Aufruf am nchsten steht (der als 
; letzter auf den Stack gepusht wird), bei Beep also die Frequenz (STDCALL-Konvention). 
; Nun kommt die berechtigte Frage: Wozu das Alles, wenn PB das selber und krzer kann? Richtig,
; fr Beep wrde man in einem PB-Programm kaum die FAsm-Version nehmen, aber PB untersttzt
; nicht direkt alle verfgbaren API-Funktionen. Die Liste der untersttzten Funktionen wird zwar
; mit jedem Versionssprung immer lnger, aber es bleiben (natrlich) noch Lcken. Schon aus
; Grnden der Kompatibilitt werden dort kaum Funktionen aufgenommen, die z.B. erst ab Windows XP
; zur Verfgung stehen. Die Liste mit den untersttzten Funktionen findet man brigens als
; Textdatei im PB-Compilers-Verzeichnis (APIFunctionListing.txt). 
; Hier jetzt ein Beispiel mit Funktionen, die PB (noch) nicht direkt untersttzt. Ermittelt werden
; die Laufwerks-Pfad-Namen der im PC vorhanden Volumes (Laufwerke). Achtung: Funktioniert erst ab
; Windows XP! 

;--------
Global Var10.l               ;wir brauchen mal wieder ne neue (Long-)Variable       

!extrn _FindFirstVolumeA@8                  ;erstes Volume finden
!extrn _FindNextVolumeA@12                  ;nchste(s) Volume(s) finden 
!extrn _FindVolumeClose@4                   ;zum ordentlichen Schliessen
!extrn _GetVolumePathNamesForVolumeNameA@16 ;ermittelt den Laufwerks-Pfad-Namen

Vol1$ = Space(1000)          ;Platz fr Ergebnis-String reservieren
Var1  = @Vol1$               ;Zeiger auf den String

Vol2$ = Space(1000)          ;Platz fr Ergebnis-String reservieren
Var2  = @Vol2$               ;Zeiger auf den String
 
Var10 = 100                  ;Lnge jeweils 100 Zeichen (berdimensioniert!)

;- Parameter auf Stack und Aufruf
!PUSH [v_Var10]              ;Lnge des Ergebnis-Strings
!PUSH [v_Var1]               ;Zeiger auf den Ergebnis-String
!CALL _FindFirstVolumeA@8    ;schreibt die Zeichen in den String Vol1$

!MOV [v_Var8],eax            ;der Rckgabe-Wert von FindFirstVolume wird als Parameter fr
                             ; FindNextVolume und FindVolumeClose bentigt
!@@:
!PUSH 0                      ;interessiert hier nicht, Null wird als Parameter akzeptiert
!PUSH [v_Var10]              ;Lnge des Ergebnis-Strings
!PUSH [v_Var2]               ;eigener Ergebnis-String
!PUSH [v_Var1]               ;Ergebnis-String von FindFirstVolume bzw. FindNextVolume
!CALL _GetVolumePathNamesForVolumeNameA@16

Vol3$ + #LFCR$+ Vol2$        ;das Ganze in Vol3$ (nicht extra definiert) speichern 

!PUSH [v_Var10]
!PUSH [v_Var1]
!PUSH [v_Var8]
!CALL _FindNextVolumeA@12

!OR eax,eax                  ;so lange suchen, bis Rchgabe-Wert von FindNextVolume Null
!jnz @b

!PUSH [v_Var8]
!CALL _FindVolumeClose@4     ;schliessen

MessageRequester("FindVolume(s)", Vol3$)
;--------
; Das "A" am Ende von Funktionen kennzeichnet diese als ANSI-Funktion, ein "W" als Unicode-
; Funktion.
; Das Ganze ist nicht sauber "programmiert", z.B. fehlt eine Auswertung von eventuellen Fehler-
; Rckmeldungen. Es sollte lediglich gezeigt werden, das es durchaus lohnenswert ist, Win32-
; API-Funktionen mit FAsm zu nutzen, wenn es mit PB (noch) nicht mglich ist.
; Details zu den oben benutzten Funktionen wrde hier den Rahmen sprengen; fr Interessierte
; gibt es das Microsoft Developer Network (MSDN).   

;=========================================================================================
; Das soll es mit dem API gewesen sein. 
; FAsm bietet auch diverse Directiven an, die das Programmieren erleichtern knnen. Im
; nachfolgenden Beispiel kombinieren wir eine Directive (virtual) mit einem Macro (writetw)
; in FAsm. Das Macro benutzt eine Directive und schreibt einen 10-Byte-Wert in den Speicher. 
; 1 Tword sind 10 Bytes und FAsm interpretiert das automatisch als 80-Bit-Float-Variable; es
; mu aber unbedingt ein Dezimal-Punkt vorhanden sein! Danach knnen weitere beliebige Ziffern
; folgen. Es mu unbedingt eine Dezimal-Zahl sein (wegen Nachkomma)! 
; Wir berechnen mal aus einer vielstelligen Float-Zahl die Quadrat-Wurzel (dazu nehmen wir die
; FPU):
;--------
Global Var11.d               ;jetzt brauchen wir mal ne Double-Variable 
Var1 = AllocateMemory(10)    ;10 Bytes fr 1 TWord 
!MOV esi,[v_Var1]            ;Zeiger auf den reservierten Speicher fr 1 TWord

;- definiert das Macro "writetw", das den bergabe-Parameter "value" erwartet
  !Macro writetw value       ;schreibt ein TWord in den Speicher
  !{virtual at 0             ;virtual ist eine FAsm-Directive!
  !  DT value                ;DT = Define TWord ;wird von FAsm automatisch als Double-Extended-
                             ; Precision-Float (=80 Bit) interpretiert
  !  local lo, mi, hi        ; 3 DWord-Variablen als "local" deklarieren; "local" = "protected"
  !  load lo dword from 0    ;die ersten 4 Byte einlesen  
  !  load mi dword from 4    ;die zweiten 4 Byte einlesen
  !  load hi word from 8     ;dann noch 2 Bytes; zusammen ist das ein TWord
  ! End virtual
  !MOV dword[esi],lo         ;die 3 lokalen Variablen in den reservierten Speicher schreiben
  !MOV dword[esi+4],mi
  !MOV word[esi+8],hi}       ;Macro-Ende
   
!writetw 12345678901234567890.0   ;das Macro schreibt den Wert des TWords in den reservierten
                                  ; Speicher, damit die FPU ihn von dort einlesen kann
;- doch ein kurzer Abstecher zur FPU
!FNINIT                      ;die FPU mit Standard-Einstellungen initialisieren  
!FLD tword[esi]              ;das TWord in die FPU einlesen
!FSQRT                       ;Quadrat-Wurzel vom TWord berechnen
!FSTP qword[v_Var11]         ;Abspeicherung in Var11
;-        

MessageRequester("Directive und Macro in FAsm"," Quadrat-Wurzel aus 12345678901234567890.0 :" + #CRLF$ + StrD(Var11))
;--------
; Was ist nun der Vorteil? Ganz einfach: Versuch mal mit PB mit so einer grossen Zahl (Stellen-
; Anzahl!) zu rechnen! Da hilft selbst die Exponential-Darstellung nicht.

End   
; IDE Options = PureBasic 4.60 Beta 3 (Windows - x86)
; CursorPosition = 529
; FirstLine = 505
; Folding = -
; EnableAsm